import os
import numpy as np

import SVAR.SVARutilGMM
import SVAR.estSVAR
import SVAR.SVARbasics
import pickle
import pytest

import SVAR.estimatorGMM
import SVAR.estimatorGMMW
import SVAR.SVARutil

np.random.seed(0)

if False:
    with open("tests\data\eps.data", 'rb') as filehandle:
        # read the data as binary data stream
        eps = pickle.load(filehandle)
    supply_eps = eps


#
# # path = os.path.join("MCResults", version)
# file_name = "W.data"
# with open(file_name, 'wb') as filehandle:
#     pickle.dump(grad, filehandle)


@pytest.fixture
def supply_eps():
    with open("data/eps.data", 'rb') as filehandle:
        # read the data as binary data stream
        eps = pickle.load(filehandle)
    return eps


@pytest.fixture
def supply_g():
    with open("data/g.data", 'rb') as filehandle:
        # read the data as binary data stream
        g = pickle.load(filehandle)
    return g


# @pytest.fixture
# def supply_W():
#     with open("data/W.data", 'rb') as filehandle:
#         # read the data as binary data stream
#         W = pickle.load(filehandle)
#     return W


@pytest.fixture
def supply_Jac():
    with open("data/Jac.data", 'rb') as filehandle:
        # read the data as binary data stream
        Jac = pickle.load(filehandle)
    return Jac


@pytest.fixture
def supply_grad():
    with open("data/grad.data", 'rb') as filehandle:
        # read the data as binary data stream
        grad = pickle.load(filehandle)
    return grad


@pytest.mark.parametrize("r,n,expected", [
    (3, 2, np.array([[1, 2], [2, 1]])),
    (2, 3, np.array([[0, 1, 1], [1, 0, 1], [1, 1, 0]])),
    (3, 3, np.array([[0, 1, 2], [0, 2, 1], [1, 0, 2], [1, 2, 0], [2, 0, 1], [2, 1, 0], [1, 1, 1]])),
    (4, 3, np.array([[0, 1, 3], [0, 3, 1], [1, 0, 3], [1, 3, 0], [3, 0, 1], [3, 1, 0], [0, 2, 2],
                     [2, 0, 2], [2, 2, 0], [1, 1, 2], [1, 2, 1], [2, 1, 1]]))
])
def test_get_Cr(r, n, expected):
    assert SVAR.SVARutilGMM.get_Cr(r, n).tolist() == expected.tolist()


@pytest.mark.parametrize("r,n,expected", [
    (3, 2, np.array([[3, 0], [0, 3]])),
    (2, 3, np.array([[2, 0, 0], [0, 2, 0], [0, 0, 2]]))
])
def test_get_Mr(r, n, expected):
    assert SVAR.SVARutilGMM.get_Mr(r, n).tolist() == expected.tolist()


@pytest.mark.parametrize("moments,expected", [
    (SVAR.SVARutilGMM.get_Cr(3, 2), np.array([[False, True, False, False, False, False, False, True, False,
                                               False],
                                              [False, False, True, False, False, False, True, False, False,
                                         False]])),
    (SVAR.SVARutilGMM.get_Cr(2, 3),
     np.array([[True, False, False, False, False, False, True, False, False,
                False, False, True, False, False, False],
               [False, True, False, False, False, True, False, False, False,
                False, False, True, False, False, False],
               [False, True, False, False, False, False, True, False, False,
                False, True, False, False, False, False]])),
    (SVAR.SVARutilGMM.get_Cr(3, 3),
     np.array([[True, False, False, False, False, False, True, False, False,
                False, False, False, True, False, False],
               [True, False, False, False, False, False, False, True, False,
                False, False, True, False, False, False],
               [False, True, False, False, False, True, False, False, False,
                False, False, False, True, False, False],
               [False, True, False, False, False, False, False, True, False,
                False, True, False, False, False, False],
               [False, False, True, False, False, True, False, False, False,
                False, False, True, False, False, False],
               [False, False, True, False, False, False, True, False, False,
                False, True, False, False, False, False],
               [False, True, False, False, False, False, True, False, False,
                False, False, True, False, False, False]])),
    (SVAR.SVARutilGMM.get_Cr(4, 3),
     np.array([[True, False, False, False, False, False, True, False, False,
                False, False, False, False, True, False],
               [True, False, False, False, False, False, False, False, True,
                False, False, True, False, False, False],
               [False, True, False, False, False, True, False, False, False,
                False, False, False, False, True, False],
               [False, True, False, False, False, False, False, False, True,
                False, True, False, False, False, False],
               [False, False, False, True, False, True, False, False, False,
                False, False, True, False, False, False],
               [False, False, False, True, False, False, True, False, False,
                False, True, False, False, False, False],
               [True, False, False, False, False, False, False, True, False,
                False, False, False, True, False, False],
               [False, False, True, False, False, True, False, False, False,
                False, False, False, True, False, False],
               [False, False, True, False, False, False, False, True, False,
                False, True, False, False, False, False],
               [False, True, False, False, False, False, True, False, False,
                False, False, False, True, False, False],
               [False, True, False, False, False, False, False, True, False,
                False, False, True, False, False, False],
               [False, False, True, False, False, False, True, False, False,
                False, False, True, False, False, False]])),
    (SVAR.SVARutilGMM.get_Mr(3, 2), np.array([[False, False, False, True, False, True, False, False, False,
                                               False],
                                              [True, False, False, False, False, False, False, False, True,
                                         False]])),
    (SVAR.SVARutilGMM.get_Mr(2, 3),
     np.array([[False, False, True, False, False, True, False, False, False,
                False, True, False, False, False, False],
               [True, False, False, False, False, False, False, True, False,
                False, True, False, False, False, False],
               [True, False, False, False, False, True, False, False, False,
                False, False, False, True, False, False]])),
])
def test_get_Moments_powerindex(moments, expected):
    assert SVAR.SVARutilGMM.get_Moments_powerindex(moments).tolist() == expected.tolist()


@pytest.mark.parametrize("moment,expected", [
    (SVAR.SVARutilGMM.get_Cr(3, 2)[0], np.array([1, 2, 2])),
    (SVAR.SVARutilGMM.get_Cr(3, 2)[1], np.array([1, 1, 2])),
    (SVAR.SVARutilGMM.get_Cr(3, 3)[0], np.array([2, 3, 3])),
    (SVAR.SVARutilGMM.get_Cr(3, 3)[1], np.array([2, 2, 3])),
    (SVAR.SVARutilGMM.get_Cr(3, 3)[6], np.array([1, 2, 3])),
    (SVAR.SVARutilGMM.get_Cr(4, 3)[0], np.array([2, 3, 3, 3])),
    (SVAR.SVARutilGMM.get_Cr(4, 3)[5], np.array([1, 1, 1, 2])),
    (SVAR.SVARutilGMM.get_Mr(2, 2)[0], np.array([1, 1])),
    (SVAR.SVARutilGMM.get_Mr(2, 2)[1], np.array([2, 2])),
    (SVAR.SVARutilGMM.get_Mr(2, 3)[0], np.array([1, 1])),
    (SVAR.SVARutilGMM.get_Mr(2, 3)[2], np.array([3, 3])),
])
def test_get_Moment_transformed(moment, expected):
    assert SVAR.SVARutilGMM.get_Moment_transformed(moment).tolist() == expected.tolist()


@pytest.mark.parametrize("b,restrictions,whiten,expected", [

    ([1, 2, 3, 4, 5, 6, 7, 8, 9], [], False,
     np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])),

    ([1, 2, 3, 4, 5, 6, 7, 8, 9],
     np.array([[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]), False,
     np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])),

    ([2, 3, 4, 5, 6, 7, 8, 9],
     np.array([[0, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]), False,
     np.array([[0, 2, 3], [4, 5, 6], [7, 8, 9]])),

    ([2, 3, 4, 5, 6, 7, 8, 9],
     np.array([[10, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]), False,
     np.array([[10, 2, 3], [4, 5, 6], [7, 8, 9]])),

    ([1, 4, 5, 7, 8, 9],
     SVAR.SVARutil.getRestrictions_recursive(np.array([[1, 0, 0], [4, 5, 0], [7, 8, 9]])), False,
     np.array([[1, 0, 0], [4, 5, 0], [7, 8, 9]])),

    ([np.pi / 2], [], True,
     np.array([[0., -1.], [1., 0.]])),

    ([0, 0, np.pi / 2], [], True,
     np.array([[1., 0., 0.], [0., 0., -1.], [0., 1., 0.]])),
])
def test_get_BMatrix(b, restrictions, whiten, expected):
    assert SVAR.SVARutil.get_BMatrix(b=b, restrictions=restrictions, whiten=whiten).tolist() == expected.tolist()


@pytest.mark.parametrize("B,restrictions,expected", [
    (np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), [], np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])),

    (np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
     np.array([[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]),
     np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])),

    (np.array([[0, 2, 3], [4, 5, 6], [7, 8, 9]]),
     np.array([[0, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]),
     np.array([2, 3, 4, 5, 6, 7, 8, 9])),

    (np.array([[10, 2, 3], [4, 5, 6], [7, 8, 9]]),
     np.array([[10, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]),
     np.array([2, 3, 4, 5, 6, 7, 8, 9])),

    (np.array([[1, 0, 0], [4, 5, 0], [7, 8, 9]]),
     SVAR.SVARutil.getRestrictions_recursive(np.array([[1, 0, 0], [4, 5, 0], [7, 8, 9]])),
     np.array([1, 4, 5, 7, 8, 9]))
])
def test_get_bVector(B, restrictions, expected):
    assert SVAR.SVARutil.get_BVector(B=B, restrictions=restrictions).tolist() == expected.tolist()


def test_get_Omega():
    # eps = supply_eps[:,:2]
    eps = np.array([[2, 3], [2, 3]])
    Omega2, Omega3, Omega4 = SVAR.SVARutil.get_Omega(eps)
    Omega2_exp = np.array([[4., 6.], [np.nan, 9.]])
    Omega3_exp = np.array([[[8., 12.],
                            [np.nan, 18.]],
                           [[np.nan, np.nan],
                            [np.nan, 27.]]])
    Omega4_exp = np.array([[[[16., 24.],
                             [np.nan, 36.]],
                            [[np.nan, np.nan],
                             [np.nan, 54.]]],
                           [[[np.nan, np.nan],
                             [np.nan, np.nan]],
                            [[np.nan, np.nan],
                             [np.nan, 81.]]]])
    assert Omega2[~np.isnan(Omega2)].tolist() == Omega2_exp[~np.isnan(Omega2)].tolist()
    assert Omega3[~np.isnan(Omega3)].tolist() == Omega3_exp[~np.isnan(Omega3)].tolist()
    assert Omega4[~np.isnan(Omega4)].tolist() == Omega4_exp[~np.isnan(Omega4)].tolist()

    assert Omega2[~np.isnan(Omega2_exp)].tolist() == Omega2_exp[~np.isnan(Omega2_exp)].tolist()
    assert Omega3[~np.isnan(Omega3_exp)].tolist() == Omega3_exp[~np.isnan(Omega3_exp)].tolist()
    assert Omega4[~np.isnan(Omega4_exp)].tolist() == Omega4_exp[~np.isnan(Omega4_exp)].tolist()


def test_get_f():
    u = np.array([[2, 3], [2, 3] ])
    b = np.array([1, 0, 0, 1])
    restrictions = []
    n = 2
    moments = SVAR.SVARutilGMM.get_Mr(2, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(3, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(4, n), axis=0)
    moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)
    whiten = False
    f = SVAR.SVARutilGMM.get_f(u, b, restrictions, moments, moments_powerindex, whiten=whiten)
    expected = np.array([[3., 8., 6., 18., 12., 54., 24., 35.],
                         [3., 8., 6., 18., 12., 54., 24., 35.]])
    assert f.tolist() == expected.tolist()

    u = np.array([[2, 3], [2, 3]])
    b = np.array([0])
    restrictions = []
    n = 2
    moments = SVAR.SVARutilGMM.get_Mr(3, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Mr(4, n), axis=0)
    moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)
    whiten = True
    f = SVAR.SVARutilGMM.get_f(u, b, restrictions, moments, moments_powerindex, whiten=whiten)
    expected = np.array([[8., 27., 13., 78.],
                         [8., 27., 13., 78.]])
    assert f.tolist() == expected.tolist()


def test_get_g(supply_eps, supply_g):
    n = 5
    u = supply_eps[:, :n]
    restrictions = []
    B = np.eye(n)
    b = SVAR.SVARutil.get_BVector(B, restrictions=restrictions)
    moments = SVAR.SVARutilGMM.get_Mr(2, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(3, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(4, n), axis=0)
    moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)
    whiten = False
    g = SVAR.SVARutilGMM.get_g(u, b, restrictions, moments, moments_powerindex, whiten)
    assert g.tolist() == supply_g.tolist()

    moments = SVAR.SVARutilGMM.get_Mr(3, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Mr(4, n), axis=0)
    moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)
    whiten = True
    b = SVAR.SVARutil.get_BVector(B, whiten=whiten)
    g = SVAR.SVARutilGMM.get_g(u, b, restrictions, moments, moments_powerindex, whiten)
    gnew = SVAR.SVARutilGMM.get_g_wf(u, b, restrictions, moments, moments_powerindex, whiten)

    g=np.round(g,5)
    gnew=np.round(gnew, 5)

    assert g.tolist() == gnew.tolist()

    moments= SVAR.SVARutilGMM.get_Moments_MIcorrection(n, blocks=False)
    moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)
    g1 = SVAR.SVARutilGMM.get_g_wf(b=b, u=u, restrictions=restrictions, moments=moments, moments_powerindex=moments_powerindex,
          whiten=whiten, blocks=False )
    g2 = SVAR.SVARutilGMM.get_g(b=b, u=u, restrictions=restrictions, moments=moments, moments_powerindex=moments_powerindex,
          whiten=whiten, blocks=False  )

    g1=np.round(g1,5)
    g2=np.round(g2, 5)

    assert g1.tolist() == g2.tolist()


def test_loss():
    u = np.array([[2, 3], [2, 3]])
    b = np.array([1, 0, 0, 1])
    restrictions = []
    n = 2
    moments = SVAR.SVARutilGMM.get_Mr(2, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(3, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(4, n), axis=0)
    moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)
    whiten = False
    g = SVAR.SVARutilGMM.get_g(u, b, restrictions, moments, moments_powerindex, whiten)
    W = np.eye(np.shape(moments)[0])
    loss = SVAR.estimatorGMM.loss(u, b, W=W, moments=moments, moments_powerindex=moments_powerindex, restrictions=restrictions)
    assert loss == 5294.0




def test_get_W_optimal(supply_eps):
    n = 2
    u = supply_eps[:, :n]
    restrictions = []
    B = np.eye(n)
    b = SVAR.SVARutil.get_BVector(B, restrictions=restrictions)
    moments = SVAR.SVARutilGMM.get_Mr(2, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(3, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(4, n), axis=0)
    moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)

    W = SVAR.SVARutilGMM.get_W_opt(u, b, restrictions, moments, Wpara='Uncorrelated')


    f = SVAR.SVARutilGMM.get_f(u=u, b=b, restrictions=restrictions, moments=moments,
              moments_powerindex=moments_powerindex)
    S = np.cov(f.T )
    W_expected = np.linalg.inv(S)

    assert W.tolist() == W_expected.tolist()

    W = SVAR.SVARutilGMM.get_W_opt(u, b, restrictions, moments, Wpara='Independent')

    e = SVAR.innovation(u, b, restrictions=restrictions)
    omega = SVAR.SVARutil.get_Omega_Moments(e)
    S = SVAR.SVARutilGMM.get_S_Indep(Moments_1=moments, Moments_2=moments, omega=omega)
    W_expected = np.linalg.inv(S)


    assert W.tolist() == W_expected.tolist()

def test_Jacobian(supply_eps, supply_Jac):
    n = 5
    u = supply_eps[:, :n]
    restrictions = np.full([n, n], np.nan)
    B = np.eye(n)
    b = SVAR.SVARutil.get_BVector(B, restrictions=restrictions)
    moments = SVAR.SVARutilGMM.get_Mr(2, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(3, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(4, n), axis=0)
    moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)

    Jacobian = SVAR.SVARutilGMM.generate_Jacobian_function(moments=moments, restrictions=restrictions)
    Omega = SVAR.SVARutil.get_CoOmega(supply_eps)
    Jac = Jacobian(u=u, b=b, restrictions=restrictions, CoOmega=Omega)

    Jac=np.round(Jac,5)
    supply_Jac=np.round(supply_Jac,5)

    assert Jac.tolist() == supply_Jac.tolist()


def test_gradient(supply_eps):
    n = 5
    u = supply_eps[:, :n]
    restrictions = np.full([n, n], np.nan)
    B = np.eye(n)
    b = SVAR.SVARutil.get_BVector(B, restrictions=restrictions)
    moments = SVAR.SVARutilGMM.get_Mr(2, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(3, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(4, n), axis=0)
    moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)
    whiten = False
    W = SVAR.SVARutilGMM.get_W_opt(u, b, restrictions, moments, Wpara='Uncorrelated')
    Jacobian = SVAR.SVARutilGMM.generate_Jacobian_function(moments=moments, restrictions=restrictions)
    grad = SVAR.estimatorGMM.gradient(u, b, Jacobian, W, restrictions, moments, moments_powerindex)
    grad = np.round(grad,8)
    grad_expected = np.array([0.04618893, -0.04448169, -0.03513598, 0.05428982, 0.06848366, -0.03515186, 0.05826086, 0.0833233, 0.06300775, 0.02590376, 0.01839882, 0.04160729, 0.09385199, -0.03826313, -0.00118679, 0.05079445, 0.08096382, -0.06218975, 0.02402831, -0.08431825, -0.02568715, 0.06836932, 0.00219649, -0.08167082, 0.07344885])

    assert grad.tolist() == grad_expected.tolist()




def test_fast_weighting(supply_eps):
    n = 4
    u = supply_eps[:, :n]

    def check_fast_weighting(b):
        # GMM_W with Wfast
        # block1 = np.array([1, 2])
        # block2 = np.array([3, 4])
        # blocks = list()
        # blocks.append(block1)
        # blocks.append(block2)
        # moments = SVAR.K_GMM.get_Moments('GMM_W', n, blocks=blocks)
        # moments_powerindex = SVAR.K_GMM.get_Moments_powerindex(moments)
        # W = SVAR.K_GMM.get_W_fast(moments)
        # J2 = SVAR.K_GMM.loss(u, b, W, restrictions=[], moments=moments, moments_powerindex=moments_powerindex, whiten=True, blocks=False)
        # #print('J2:', J2)

        # GMM_W with Wfast
        moments = SVAR.SVARutilGMM.get_Cr(4, 4)
        moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)
        W = SVAR.estimatorGMMW.get_W_fast(moments)
        J = SVAR.estimatorGMMW.loss(u, b, W, restrictions=[], moments=moments, moments_powerindex=moments_powerindex,
                                     blocks=False)
        # print('J:', J)

        # GMM fast
        moments = SVAR.SVARutilGMM.get_Mr(4, 4)
        moments_powerindex = SVAR.SVARutilGMM.get_Moments_powerindex(moments)
        W = np.eye(np.shape(moments)[0])
        H = SVAR.estimatorGMMW.loss(u, b, W, restrictions=[], moments=moments, moments_powerindex=moments_powerindex,
                                      blocks=False)
        # print('H:', H)

        # print('J+H',J+H)
        # print('J2+H',J2+H)
        # print(" ")
        return J, H

    b = np.array([1., 0, 0., 0., 0., 0.])
    J1, H1 = check_fast_weighting(b)

    b = np.array([0., 1, 0., 0., 0., 0.])
    J2, H2 = check_fast_weighting(b)

    b = np.array([0., 1, 1., 1., 0., 1.])
    J3, H3 = check_fast_weighting(b)

    assert np.round(J1 + H1,5) == np.round(J2 + H2,5)
    assert np.round(J1 + H1,5) == np.round(J3 + H3,5)
    assert np.round(J2 + H2,5) == np.round(J3 + H3,5)



import numpy as np
import SVAR


def get_B_Cholesky(u):
    T, n = np.shape(u)
    Sigma = np.dot(np.transpose(u), u) / T
    B = np.linalg.cholesky(Sigma)
    e = np.transpose(np.matmul(np.linalg.inv(B), np.transpose(u)))
    Omega = SVAR.SVARutil.get_Omega(e)

    out = dict()
    out['B_est'] = B
    out['b_est'] = SVAR.get_BVector(B)
    out['e'] = e
    out['loss'] = 0
    out['Omega_all'] = Omega
    out['options'] = dict()
    return out


from scipy import optimize as opt 
import numpy as np
import pandas as pd
import SVAR
import SVAR.SVARutilGMM
import SVAR.estOutput
import SVAR.estimatorPML
import SVAR.estimatorGMM
import SVAR.estimatorGMMW
import SVAR.estimatorGMMWF
import SVAR.estimatorCholesky
from scipy import optimize
import copy
from scipy.optimize import fmin_slsqp

def SVARest(u, estimator='GMM', options=dict(), prepOptions=dict(), prepared=False):
    if estimator == 'GMM':
        if not (prepared):
            prepOptions['estimator'] = 'GMM'
            options = SVAR.estimatorGMM.prepareOptions(u=u, **prepOptions)

        def optimize_this(options):
            this_loss = lambda b_vec: SVAR.estimatorGMM.loss(u, b_vec,
                                                             restrictions=options['restrictions'],
                                                             moments=options['moments'],
                                                             moments_powerindex=options['moments_powerindex'],
                                                             W=options['W'] )
            # this_grad = lambda b_vec: SVAR.estimatorGMM.gradient(u, b_vec,
            #                                                      Jacobian=options['Jacobian'],
            #                                                      W=options['W'],
            #                                                      restrictions=options['restrictions'],
            #                                                      moments=options['moments'],
            #                                                      moments_powerindex=options['moments_powerindex'])
            this_grad = []
            optim_start = options['bstart']
            if np.shape(optim_start)[0] == 0:
                # raise ValueError('No free parameters.')
                ret = dict()
                ret['x'] = np.array([])
                ret['fun'] = this_loss(ret['x'])
            else:

                ret_tmp = opt.minimize(this_loss, optim_start, method='L-BFGS-B', jac=this_grad,
                             bounds=None,
                             tol=None,
                             callback=None,
                             options={'disp': False,
                                      'maxcor': 10,
                                      'ftol': 2.220446049250313e-09,
                                      'gtol': 1e-09,
                                      'eps': 1e-12,
                                      'maxfun': 15000,
                                      'maxiter': 15000,
                                      'iprint': - 1,
                                      'maxls': 40,
                                      'finite_diff_rel_step': None})

                if ret_tmp['success']==False:
                    print('Optimization failed')
                    print(ret_tmp['message'])
                ret = dict()
                ret['x'] = ret_tmp.x
                ret['fun'] = ret_tmp.fun

                if ret['fun'] < 0:
                    raise ValueError("Error: Negative loss")
            return ret

        est_SVAR = optimize_this(options)

        for k in range(1, options['kstep']):
            bstart = est_SVAR['x']
            options['W'] = SVAR.SVARutilGMM.get_W_opt(u, b=bstart, restrictions=options['restrictions'],
                                                      moments=options['moments'],
                                                      Wpara=options['Wpara'],
                                                      S_func=options['S_func'])
            est_SVAR = optimize_this(options)


        out_SVAR = SVAR.estimatorGMM.SVARout(est_SVAR, options, u)

    elif estimator =='CSUE':
        if not (prepared):
            prepOptions['estimator'] = 'CSUE'
            options = SVAR.estimatorGMM.prepareOptions(u=u, **prepOptions)

        def optimize_this(options):

            get_d = lambda b_vec: np.divide(1,np.sqrt(np.sum(np.power(SVAR.innovation(u, b_vec ), 2),axis=0)/ options['T']))
            get_dprod = lambda b_vec: np.prod(np.power(get_d(b_vec), options['moments'] ),axis=1)


            Wupd = lambda b_vec: np.matmul(np.matmul(  np.diag( get_dprod(b_vec)  ) ,options['W']), np.diag( get_dprod(b_vec)  ) )

            this_loss = lambda b_vec: SVAR.estimatorGMM.loss(u, b_vec,
                                                             restrictions=options['restrictions'],
                                                             moments=options['moments'],
                                                             moments_powerindex=options['moments_powerindex'],
                                                             W=Wupd(b_vec) )

            this_grad = []

            optim_start = options['bstart']
            if np.shape(optim_start)[0] == 0:
                # raise ValueError('No free parameters.')
                ret = dict()
                ret['x'] = np.array([])
                ret['fun'] = this_loss(ret['x'])
            else:
                ret_tmp = opt.minimize(this_loss , optim_start, method='L-BFGS-B', jac=this_grad )
                if ret_tmp['success'] == False:
                    print('Optimization failed')
                    print(ret_tmp['message'])
                ret = dict()
                ret['x'] = ret_tmp.x
                ret['fun'] = ret_tmp.fun

            return ret

        est_SVAR = optimize_this(options)


        for k in range(1, options['kstep']):
            bstart = est_SVAR['x']
            options['W'] = SVAR.SVARutilGMM.get_W_opt(u, b=bstart, restrictions=options['restrictions'],
                                                      moments=options['moments'],
                                                      Wpara=options['Wpara'],
                                                      S_func=options['S_func'])
            est_SVAR = optimize_this(options)

        out_SVAR = SVAR.estimatorGMM.SVARout(est_SVAR, options, u)

    elif estimator == 'CUE':
        if not (prepared):
            prepOptions['estimator'] = 'CUE'
            options = SVAR.estimatorGMM.prepareOptions(u=u, **prepOptions)

        def optimize_this(options):
            this_loss = lambda b_vec: SVAR.estimatorGMM.loss(u, b_vec,
                                                             restrictions=options['restrictions'],
                                                             moments=options['moments'],
                                                             moments_powerindex=options['moments_powerindex'],
                                                             W=SVAR.SVARutilGMM.get_W_opt(u, b=b_vec,
                                                                                          restrictions=options[
                                                                                               'restrictions'],
                                                                                          moments=options['moments'],
                                                                                          Wpara=options['Wpara'],
                                                                                          S_func=options['S_func']))




            #eps = np.sqrt(np.finfo(float).eps)
            #this_grad = lambda b_vec: optimize.approx_fprime(b_vec, this_loss, eps)
            this_grad = []

            optim_start = options['bstart']
            if np.shape(optim_start)[0] == 0:
                # raise ValueError('No free parameters.')
                ret = dict()
                ret['x'] = np.array([])
                ret['fun'] = this_loss(ret['x'])
            else:
                # ret_tmp = opt.minimize(this_loss, optim_start, method='L-BFGS-B', jac=this_grad,options={'disp': True})
                ret_tmp = opt.minimize(this_loss, optim_start, method='L-BFGS-B', jac=this_grad )
                if ret_tmp['success']==False:
                    print('Optimization failed')
                    print(ret_tmp['message'])
                ret = dict()
                ret['x'] = ret_tmp.x
                ret['fun'] = ret_tmp.fun


                if ret['fun'] < 0:
                    raise ValueError("Error: Negative loss")
            return ret

        est_SVAR = optimize_this(options)

        options['W'] = SVAR.SVARutilGMM.get_W_opt(u, b=est_SVAR['x'], restrictions=options['restrictions'],
                                                  moments=options['moments'],
                                                  Wpara=options['Wpara'],
                                                  S_func=options['S_func'])
        out_SVAR = SVAR.estimatorGMM.SVARout(est_SVAR, options, u)

    elif estimator == 'GMM_W':
        if not (prepared):
            prepOptions['estimator'] = 'GMM_W'
            options = SVAR.estimatorGMMW.prepareOptions(u=u, **prepOptions)
        z, options['V'] = SVAR.do_whitening(u, white=True)

        def optimize_this(options):
            this_loss = lambda b_vec: SVAR.estimatorGMMW.loss(z, b_vec,
                                                              restrictions=options['restrictions'],
                                                              moments=options['moments'],
                                                              moments_powerindex=options['moments_powerindex'],
                                                              W=options['W'],
                                                              blocks=options['blocks'])
            #this_grad =  lambda b_vec: optimize.approx_fprime(b_vec, this_loss, epsilon=1.4901161193847656e-08)
            this_grad =[]

            optim_start = options['bstart']
            if np.shape(optim_start)[0] == 0:
                # raise ValueError('No free parameters.')
                ret = dict()
                ret['x'] = np.array([])
                ret['fun'] = this_loss(ret['x'])
            else:
                ret_tmp = opt.minimize(this_loss, optim_start, method='L-BFGS-B', jac=this_grad)
                if ret_tmp['success']==False:
                    print('Optimization failed')
                    print(ret_tmp['message'])
                ret = dict()
                ret['x'] = ret_tmp.x
                ret['fun'] = ret_tmp.fun

                if ret['fun'] < 0:
                    raise ValueError("Error: Negative loss")

            return ret

        est_SVAR = optimize_this(options)

        out_SVAR = SVAR.estimatorGMMW.SVARout(est_SVAR, options, u)

    elif estimator == 'GMM_WF':
        if not (prepared):
            prepOptions['estimator'] = 'GMM_WF'
            options = SVAR.estimatorGMMWF.prepareOptions(u=u, **prepOptions)
        z, options['V'] = SVAR.do_whitening(u, white=True)

        def optimize_this(options):
            if options['moments_MeanIndep']:
                this_loss = lambda b_vec: SVAR.estimatorGMMWF.loss(z, b_vec,
                                                                   restrictions=options['restrictions'],
                                                                   moments=options['moments'],
                                                                   moments_powerindex=options['moments_powerindex'],
                                                                   blocks=options['blocks'])-\
                                        SVAR.estimatorGMMWF.loss_MIcorrection(z,b_vec,restrictions=options['restrictions'],
                                                                  moments=options['moments_MIcorrection'],
                                                                  moments_powerindex=options['moments_MIcorrection_powerindex'], blocks=options['blocks'])
            else:
                this_loss = lambda b_vec: SVAR.estimatorGMMWF.loss(z, b_vec,
                                                                   restrictions=options['restrictions'],
                                                                   moments=options['moments'],
                                                                   moments_powerindex=options['moments_powerindex'],
                                                                   blocks=options['blocks'])
            #this_grad =  lambda b_vec: optimize.approx_fprime(b_vec, this_loss, epsilon=1.4901161193847656e-08)
            this_grad =[]

            optim_start = options['bstart']
            if np.shape(optim_start)[0] == 0:
                # raise ValueError('No free parameters.')
                ret = dict()
                ret['x'] = np.array([])
                ret['fun'] = this_loss(ret['x'])
            else:

                ret_tmp = opt.minimize(this_loss, optim_start, method='L-BFGS-B', jac=this_grad)

                if ret_tmp['success'] == False:
                    print('Optimization failed')
                    print(ret_tmp['message'])
                ret = dict()
                ret['x'] = ret_tmp.x
                ret['fun'] = ret_tmp.fun
                # if ret['fun'] < 0:
                #     raise ValueError("Error: Negative loss")
            return ret

        est_SVAR = optimize_this(options)

        out_SVAR = SVAR.estimatorGMMWF.SVARout(est_SVAR, options, u)

    elif estimator == 'PML':
        if not (prepared):
            prepOptions['estimator'] = 'PML'
            options = SVAR.estimatorPML.prepareOptions(u=u, **prepOptions)
        z, options['V'] = SVAR.do_whitening(u, white=True)

        def optimize_this(options):
            this_loss = lambda b_vec: SVAR.estimatorPML.LogLike_t(z, b_vec, options['df_t'], blocks=options['blocks'],
                                                                  whiten=options['whiten'])
            this_grad = []

            optim_start = options['bstart']
            if np.shape(optim_start)[0] == 0:
                # raise ValueError('No free parameters.')
                ret = dict()
                ret['x'] = np.array([])
                ret['fun'] = this_loss(ret['x'])
            else:
                ret_tmp = opt.minimize(this_loss, optim_start, method='L-BFGS-B', jac=this_grad)
                if ret_tmp['success'] == False:
                    print('Optimization failed')
                    print(ret_tmp['message'])
                ret = dict()
                ret['x'] = ret_tmp.x
                ret['fun'] = ret_tmp.fun


            return ret

        est_SVAR = optimize_this(options)

        out_SVAR = SVAR.estimatorPML.SVARout(est_SVAR, options, u)

    elif estimator == 'Cholesky':
        out_SVAR = SVAR.estimatorCholesky.get_B_Cholesky(u)

    else:
        print('Unknown estimator')


    return out_SVAR



import numpy as np
import SVAR.SVARbasics
import SVAR.estSVAR
import SVAR.estPrepare
from SVAR.estPrepare import prepare_bootstrapIRF
import pickle
import pathos
import copy
mp = pathos.helpers.mp


def bootstrap_estimation(y,  estimator='GMM', opt_bootstrap=dict(), opt_redform=dict(),
                         prepOptions=dict(),
                         opt_SVAR=dict(), opt_irf=dict(), prepared=False, Bfirst=[]):
    out_redform = SVAR.SVARbasics.OLS_ReducedForm(y, **opt_redform)



    # Estimate B
    if prepared:
        opt_SVAR['printOutput'] = False
        out_svar = SVAR.SVARest(out_redform['u'], estimator=estimator, options=opt_SVAR, prepared=True)
    else:
        prepOptions['printOutput'] = True
        out_svar = SVAR.SVARest(out_redform['u'], estimator=estimator, prepOptions=prepOptions)

    # Find permutation
    if opt_bootstrap['find_perm'] == True:
        if Bfirst != []:
            try:
                # Find best permutation
                T_this=np.shape(out_redform['u'])[0]
                V = out_svar['Avar_est']
                V = V[np.isnan(V) == False].reshape((np.sum(np.isnan(out_svar['options']['restrictions'])),
                                                     np.sum(np.isnan(out_svar['options']['restrictions']))))
                [Bbest, permutation] = SVAR.SVARutil.PermToB0(Bfirst, out_svar['B_est'], V, out_svar['options']['restrictions'],
                                                              T_this)
                est_SVAR = dict()

                if  out_svar['options']['whiten']:
                    Bbest = np.matmul(np.linalg.inv( out_svar['options']['V']), Bbest)

                est_SVAR['x'] = SVAR.get_BVector(Bbest, out_svar['options']['restrictions'], whiten=out_svar['options']['whiten'],blocks=out_svar['options']['blocks'])

                if estimator == 'CUE' or estimator == 'CSUE':
                    out_svar = SVAR.estimatorGMM.SVARout(est_SVAR, out_svar['options'], out_redform['u'])
                elif estimator == 'GMM_W':
                    out_svar = SVAR.estimatorGMMW.SVARout(est_SVAR, out_svar['options'], out_redform['u'])
                elif estimator == 'GMM_WF':
                    out_svar = SVAR.estimatorGMMWF.SVARout(est_SVAR, out_svar['options'], out_redform['u'])
                elif estimator == 'PML':
                    out_svar = SVAR.estimatorPML.SVARout(est_SVAR, out_svar['options'], out_redform['u'])
                else:
                    print("Bootstrap find best permutation unknown estimator")
            except:
                print("Boostrap could not find best permutation")



    # get IRF
    out_irf = SVAR.SVARbasics.get_IRF(out_svar['B_est'], out_redform['AR'],   **opt_irf)

    return out_redform, out_svar, out_irf, out_svar['options']


def bootstrap_SVAR(y, estimator='GMM',
                   options_bootstrap=dict(), options_redform=dict(), prepOptions=dict(), options_irf=dict(),
                    estimatorname = ''):
    # Bundle inputs to options
    options_bootstrap, options_redform, options_irf = prepare_bootstrapIRF(**options_bootstrap, **options_redform,
                                                                   **options_irf)

    parallel = options_bootstrap['parallel']
    num_cores_boot = options_bootstrap['num_cores_boot']
    seed_start = options_bootstrap['seed_start']
    saveout = options_bootstrap['saveout']


    out_redform, out_svar, out_irf, opt_SVAR = bootstrap_estimation(y,  estimator,
                                                                    opt_bootstrap=options_bootstrap,
                                                                    opt_redform=options_redform,
                                                                    prepOptions=prepOptions, opt_SVAR=[],
                                                                    opt_irf=options_irf,
                                                                    prepared=False )


    # save first Estimation
    if saveout:
        out_svar_save = copy.deepcopy(out_svar)
        out_svar_save['options']['Jacobian']=np.nan
        out_svar_save['options']['S_func'] = np.nan
        out_svar_save['options']['Sdel_func'] = np.nan
        file_name = options_bootstrap['path'] + "/" + estimatorname + "/EstimationFirst/" + str(
            seed_start) + ".data"
        with open(file_name, 'wb') as filehandle:

            pickle.dump([out_redform, out_svar_save], filehandle)

        # save first IRF
        file_name = options_bootstrap['path'] + "/" + estimatorname + "/IRFFirst/" + str(
            seed_start) + ".data"
        with open(file_name, 'wb') as filehandle:
            pickle.dump(out_irf, filehandle)




    if num_cores_boot > 1:
        parallel = True


    def bootstrap_iter(y, estimator, options_bootstrap, options_redform, prepOptions, opt_SVAR,
                       options_irf, out_redform, jobno=0, seed_start=0, estimatorname='', Bfirst=[]):


        y_sim = simulate_SVAR(y,**out_redform, jobno=jobno, seed_start=seed_start)

        out_redform_iter, out_svar_iter, out_irf_iter, _ = bootstrap_estimation(y_sim,  estimator,
                                                                                options_bootstrap,
                                                                                options_redform, prepOptions,
                                                                                opt_SVAR, options_irf,
                                                                                prepared=True,
                                                                                Bfirst=Bfirst)

        # Save results
        if saveout:
            file_name = options_bootstrap['path'] +"/" + estimatorname  +"/IRFs/Bootstrap_" + str(
                seed_start) + 'run_' + str(jobno) + ".data"
            with open(file_name, 'wb') as filehandle:
                pickle.dump(out_irf_iter, filehandle)

        return out_redform_iter['AR'], out_svar_iter['B_est'], out_irf_iter


    if parallel:
        poolBoot = mp.Pool(num_cores_boot)  # ProcessPool #mp.Pool(num_coresCV)
        result_boot_async = poolBoot.starmap_async(bootstrap_iter, [(y, estimator, options_bootstrap,
                                                                     options_redform, prepOptions, opt_SVAR,
                                                                     options_irf, out_redform, bootstrap_run,
                                                                     seed_start,
                                                                     estimatorname,out_svar['B_est'])
                                                                    for bootstrap_run in
                                                                    range(options_bootstrap['number_bootstrap'])])
        result_boot = result_boot_async.get()

        poolBoot.close()
        poolBoot.join()
        out_bootstrap = result_boot

    else:

        # Loop through iterations
        out_bootstrap = [bootstrap_iter(y, estimator, options_bootstrap,
                                            options_redform, prepOptions, opt_SVAR, options_irf, out_redform,
                                            bootstrap_run, seed_start, estimatorname,out_svar['B_est']) for bootstrap_run in
                             range(options_bootstrap['number_bootstrap'])]


    return out_irf, out_bootstrap, out_svar['B_est'], out_redform['AR'],





def simulate_SVAR(y, u, AR, const, trend, trend2, jobno=1, seed_start=0, burn_in=500):
    np.random.seed(seed_start + jobno)
    T, n = u.shape
    lags = AR.shape[2]

    # Pre-allocate the entire array
    y_new = np.zeros((T + lags + burn_in, n))
    y_new[:lags] = y[:lags]

    # Resample u
    u_resample = u[np.random.choice(T, T + lags + burn_in, replace=True)]

    # Pre-compute trend terms
    trend_terms = np.outer(np.arange(-lags, T + burn_in - lags), trend) + \
                  np.outer(np.arange(-lags, T + burn_in - lags)**2, trend2)

    for t in range(lags, T + lags + burn_in):
        y_new[t] = const + trend_terms[t-lags] + u_resample[t-lags]
        for j in range(lags):
            y_new[t] += AR[:, :, j] @ y_new[t - j - 1]

    return y_new[lags + burn_in:]

def simulate_SVAR2(e,B, AR, const, trend, trend2,ystart):
    T, n = np.shape(e)
    lags = np.shape(AR)[2]



    y_new = np.zeros([T, n])
    for t in range(T):
        tmpsum = const + np.multiply(trend, t - lags) + np.multiply(trend2, t - lags)
        for j in range(lags):
            if t - j > 0:
                tmpsum = tmpsum + np.matmul(AR[:, :, j], y_new[t - j - 1])
            else:
                tmpsum = tmpsum + np.matmul(AR[:, :, j], ystart[t-j-1])
        tmpsum = tmpsum + np.matmul(B,e[t,:])
        y_new[t, :] = tmpsum

    return y_new


import numpy as np
import SVAR
import SVAR.SVARutilGMM
import SVAR.estimatorPML
import SVAR.estimatorGMM
from SVAR.estimatorCholesky import get_B_Cholesky
from SVAR.SVARutil import get_BVector, get_block_rec, getRestrictions_blocks


# Prepare Boostrap
def prepare_bootstrapIRF(number_bootstrap=500, update_bstart=False, find_perm=True, path ='',
                         lags=1, add_const=True, add_trend=False, add_trend2=False,
                         irf_length=12,scale=False,parallel = False, num_cores_boot = 1,
                         seed_start = 0,   saveout = True  ):
    opt_bootstrap = dict()
    opt_bootstrap['number_bootstrap'] = number_bootstrap
    opt_bootstrap['update_bstart'] = update_bstart
    opt_bootstrap['find_perm'] = find_perm
    opt_bootstrap['path'] = path
    opt_bootstrap['parallel'] = parallel
    opt_bootstrap['num_cores_boot'] = num_cores_boot
    opt_bootstrap['seed_start'] = seed_start
    opt_bootstrap['saveout'] = saveout



    opt_redform = dict()
    opt_redform['lags'] = lags
    opt_redform['add_const'] = add_const
    opt_redform['add_trend'] = add_trend
    opt_redform['add_trend2'] = add_trend2

    opt_irf = dict()
    opt_irf['irf_length'] = irf_length
    opt_irf['scale'] = scale

    return opt_bootstrap, opt_redform, opt_irf


# Prepare opt_SVAR functions
def prepare_W(u, n, W, Wstartopt, moments, bstart, restrictions,   Wpara, S_func=False  ):
    if not (Wstartopt == 'I' or Wstartopt == 'WoptBstart' or Wstartopt == 'WoptparaNorm' or Wstartopt == 'specific'):
        raise ValueError('Unknown SVAR option Wstartopt.')
    if Wstartopt == 'I':
        W = np.eye(np.shape(moments)[0])
    elif Wstartopt == 'WoptBstart':
        W = SVAR.SVARutilGMM.get_W_opt(u, bstart, restrictions, moments,
                                       Wpara=Wpara, S_func=S_func)
    elif Wstartopt == 'WoptparaNorm':
        omega_normdist = np.array([0, 1, 0, 3, 0, 15])  # moments of normal distribution
        omega_normdist = np.tile(omega_normdist, (n, 1))
        W = np.linalg.inv(SVAR.SVARutilGMM.get_S_Indep(Moments_1=moments, Moments_2=moments, omega=omega_normdist ) )

    return W

def prepare_blocks_restrictions(n, n_rec, blocks, restrictions):
    # if no restrictions -> set options to no restrictions
    if not (n_rec) and blocks == False and np.array(restrictions).size == 0:
        restrictions = np.full([n, n], np.nan)
        # opt_SVAR['blocks'] == False

    # if n_rec specified -> overwrite blocks
    if not (n_rec == False):
        blocks = get_block_rec(n_rec, n)

    # if blocks specified -> overwrite restrictions
    if not (blocks == False):
        restrictions = getRestrictions_blocks(blocks)

    return restrictions, blocks

def prepare_moments(estimator, moments, addThirdMoments, addFourthMoments, moments_blocks, blocks, n, moments_MeanIndep=False):
    if np.array(moments).size == 0:
        if moments_blocks:
            blocks = blocks
        else:
            blocks = False
        moments = SVAR.SVARutilGMM.get_Moments(estimator, n, blocks=blocks,
                                               addThirdMoments=addThirdMoments,
                                               addFourthMoments=addFourthMoments, moments_MeanIndep=moments_MeanIndep)
    return moments

def prepare_bstart(estimator, bstart, u, options, bstartopt='Rec' ):
    if estimator == 'GMM' or estimator == 'CUE' or estimator == 'CSUE':
        if not (bstartopt == 'Rec' or bstartopt == 'CUEN' or bstartopt == 'I' or bstartopt == 'GMM_WF' or bstartopt == 'specific'):
            raise ValueError('Unknown SVAR option bstartopt.')
        if bstartopt == 'Rec':
            out = get_B_Cholesky(u)
            B_start = out['B_est']
            bstart = get_BVector(B_start, restrictions=options['restrictions'])
        if bstartopt == 'CUEN':
            T, n = np.shape(u)
            prepOptions_this = dict()
            prepOptions_this['printOutput'] = False
            prepOptions_this['blocks'] = options['blocks']
            prepOptions_this['restrictions'] = options['restrictions']
            prepOptions_this['moments'] = SVAR.SVARutilGMM.get_Moments('GMM', n, blocks=options['blocks'],
                                                         addThirdMoments=True,
                                                         addFourthMoments=True,
                                                         moments_MeanIndep=options['moments_MeanIndep'])
            prepOptions_this['Wpara'] = 'Independent'
            prepOptions_this['Avarparametric'] = 'Independent'
            prepOptions_this['Wstartopt'] = 'WoptBstart'
            prepOptions_this['bstartopt'] = 'GMM_WF'
            out = SVAR.SVARest(u, estimator='CUE', prepOptions=prepOptions_this)
            B_start = out['B_est']
            bstart = get_BVector(B_start, restrictions=options['restrictions'])
        elif bstartopt == 'I':
            n = np.shape(u)[1]
            B_start = np.eye(n)
            bstart = get_BVector(B_start, restrictions=options['restrictions'])
        elif bstartopt == 'GMM_WF':
            prepOptions_this = dict()
            prepOptions_this['printOutput'] = False
            prepOptions_this['blocks'] = options['blocks']
            out = SVAR.SVARest(u, estimator='GMM_WF', prepOptions=prepOptions_this)
            B_start = out['B_est']
            bstart = get_BVector(B_start, restrictions=options['restrictions'])

    if estimator == 'GMM_W' or estimator == 'GMM_WF' or estimator == 'PML' or estimator == 'RidgeBWF':
        T, n = np.shape(u)
        if not (bstartopt == 'Rec' or bstartopt == 'specific' or bstartopt == 'GMM_WF'):
            raise ValueError('Unknown SVAR option bstartopt.')
        if bstartopt == 'Rec':
            if options['blocks']:
                b_counter = 0
                for block in options['blocks']:
                    n = block[1] - block[0] + 1
                    il1 = np.tril_indices(n, k=-1)
                    n_s = int(np.size(il1) / 2)
                    b_counter = b_counter + n_s
                bstart = np.zeros(b_counter)
            else:
                bstart = np.zeros(int(n * (n - 1) / 2))
        if bstartopt == 'GMM_WF':
            prepOptions_this = dict()
            prepOptions_this['printOutput'] = False
            prepOptions_this['blocks'] = options['blocks']
            out = SVAR.SVARest(u, estimator='GMM_WF', prepOptions=prepOptions_this)
            bstart = out['b_est']


    return bstart


import numpy as np
import SVAR


def LogLike_t(u, b_vec, df_t, blocks=False,whiten=True):
    e = SVAR.SVARutil.innovation(u, b_vec, whiten=whiten, blocks=blocks)
    loglike = np.sum((-1) * ((1 - df_t) / 2) * np.log(1 + np.power(e, 2) / (df_t - 2)))
    return loglike

def PML_avar(n):
    # ToDo: V_est for PML
    V_est = np.full([n * n, n * n], np.nan)
    return V_est

# def LogLike_t_approx(u, b_vec, df_t, blocks=False,whiten=True):
#     e = SVAR.utilities.innovation(u, b_vec, whiten=whiten, blocks=blocks)
#     loglike = np.sum(((1 - df_t) / (2 * 2 * (df_t - 2))) * np.power(e, 4))
#     return loglike
#
# def LogLike_t_approx2(u, b_vec, df_t, blocks=False,whiten=True):
#     e = SVAR.utilities.innovation(u, b_vec, whiten=whiten, blocks=blocks)
#     loglike = np.sum(((1 - df_t) / (2 * 2 * (df_t - 2))) * np.power(e, 4) +
#                      ((1 - df_t) / (2 * 3 * (df_t - 2))) * np.power(e, 6) )
#     return loglike

def prepareOptions(u,
                   df_t = 7,
                   bstart=[], bstartopt='Rec',
                   blocks=False, n_rec=False,
                   printOutput=True,
                    estimator = 'PML'
                   ):
    options = dict()


    options['estimator'] = estimator

    T, n = np.shape(u)
    options['T'] = T
    options['n'] = n

    options['df_t'] = df_t

    options['printOutput'] = printOutput

    options['whiten'] = True
    _, V = SVAR.do_whitening(u, white=True)
    options['V'] = V

    restrictions, blocks = SVAR.estPrepare.prepare_blocks_restrictions(n, n_rec, blocks, restrictions=[])
    options['restrictions'] = restrictions
    options['blocks'] = blocks


    bstart = SVAR.estPrepare.prepare_bstart(estimator, bstart, u, options, bstartopt=bstartopt)
    options['bstart'] = bstart

    return options


def SVARout(est_SVAR, options, u):
    T, n = np.shape(u)
    out_SVAR = dict()
    out_SVAR['options'] = options

    b_est = est_SVAR['x']
    out_SVAR['b_est'] = b_est

    B_est = SVAR.get_BMatrix(b_est, restrictions=options['restrictions'], whiten=options['whiten'],
                             blocks=options['blocks'])
    B_est = np.matmul(options['V'], B_est)
    out_SVAR['B_est'] = B_est

    e = SVAR.innovation(u, SVAR.get_BVector(B_est, restrictions=options['restrictions']),
                        restrictions=options['restrictions'],
                        blocks=options['blocks'])
    out_SVAR['e'] = e

    Omega_all = SVAR.SVARutil.get_Omega(e)
    out_SVAR['Omega_all'] = Omega_all
    omega = SVAR.SVARutil.get_Omega_Moments(e)
    out_SVAR['omega'] = omega

    out_SVAR['loss'] = est_SVAR['fun']

    V_est = PML_avar(options['n'])
    out_SVAR['Avar_est'] = V_est

    if options['printOutput']:
        SVAR.estOutput.print_out(n, T, out_SVAR)

    return out_SVAR















import numpy as np
import SVAR
import SVAR.SVARutilGMM
from SVAR.SVARutilGMM import  get_S_uncentered, get_g, get_G_Indep, get_S_Indep, get_S
import scipy.stats


def loss(u, b, W, restrictions, moments, moments_powerindex, blocks=False ):
    whiten = False
    g = get_g(b=b, u=u, restrictions=restrictions, moments=moments, moments_powerindex=moments_powerindex,
              whiten=whiten, blocks=blocks )
    Q = np.linalg.multi_dot([g, W, g])


    return Q


def gradient(u, b, Jacobian, W, restrictions, moments, moments_powerindex):
    e = SVAR.innovation(u, b, restrictions=restrictions)
    CoOmega = SVAR.SVARutil.get_CoOmega(e)

    Jac_temp = Jacobian(u=u, b=b, restrictions=restrictions, CoOmega=CoOmega)
    this_g = get_g(b=b[:], u=u[:], restrictions=restrictions[:],
                   moments=moments[:], moments_powerindex=moments_powerindex)
    dGMM_temp  =  2 * np.linalg.multi_dot([np.transpose(Jac_temp), W, this_g])
    return (dGMM_temp)


def gradient_cont(u, b, Jacobian, W, restrictions, moments, moments_powerindex, Sdel_func):
    n = np.shape(moments)[1]
    B = SVAR.get_BMatrix(b, restrictions=restrictions)
    A = np.linalg.inv(B)
    e = SVAR.innovation(u, b, restrictions=restrictions)
    omegaext = SVAR.SVARutil.get_omegaext(e)
    CoOmega = SVAR.SVARutil.get_CoOmega(e)

    Jac_temp = Jacobian(u=u, b=b, restrictions=restrictions, CoOmega=CoOmega)
    this_g = get_g(b=b[:], u=u[:], restrictions=restrictions[:],
                   moments=moments[:], moments_powerindex=moments_powerindex)
    dGMM_temp = 2 * np.linalg.multi_dot([np.transpose(Jac_temp), W, this_g])

    counter = 0
    for q in range(n):
        for l in range(n):
            # only if B element is not restricted
            if np.isnan(restrictions[q, l]):
                L = Sdel_func[counter](omegaext,A)
                dGMM_temp2 = np.linalg.multi_dot([np.transpose(this_g), W, L, W, this_g])
                dGMM_temp[counter] = dGMM_temp[counter] + dGMM_temp2
                counter = counter + 1

    return (dGMM_temp)


# Estimate Avar and tests
def get_Avar(n, G,S,W=[],restrictions=[]):
    if np.array(W).size == 0:
        W = np.linalg.inv(S)

    if np.array(restrictions).size == 0:
        restrictions = np.full([n, n], np.nan)

    M1 = np.linalg.inv(np.matmul(np.matmul(np.transpose(G), W), G))
    M2 = np.matmul(np.transpose(G), W)
    M = np.matmul(M1, M2)
    V_est = np.matmul(np.matmul(M, S), np.transpose(M))

    elementcounter = 0
    for i in range(n):
        for j in range(n):
            if not (np.isnan(restrictions[i, j])):
                V_est = np.insert(V_est, elementcounter, np.full(np.shape(V_est)[1], np.NaN), 0)
                V_est = np.insert(V_est, elementcounter, np.full(np.shape(V_est)[0], np.NaN), 1)
            elementcounter += 1
    return V_est




# Only for GMM (not w or wf)
def prepareOptions(u,
                   addThirdMoments=True, addFourthMoments=True, moments=[], moments_blocks=True,  moments_MeanIndep=False,
                   bstart=[], bstartopt='Rec',
                   Avarparametric='Uncorrelated',
                   restrictions=[], blocks=False, n_rec=False,
                   kstep=2, W=[], Wpara='Uncorrelated', Wstartopt='I',   S_func=False,
                   printOutput=True,
                     WupdateInOutput=False,
                   estimator='GMM'):
    options = dict()

    options['estimator'] = estimator

    T, n = np.shape(u)
    options['T'] = T
    options['n'] = n


    options['WupdateInOutput'] = WupdateInOutput

    options['printOutput'] = printOutput

    if Avarparametric == 'Uncorrelated' or Avarparametric == 'Uncorrelated_uncentered'  or Avarparametric == 'Independent':
        options['Avarparametric'] = Avarparametric
    else:
        print("Invalid Avarparametric value. Set Avarparametric=Uncorrelated")
        options['Avarparametric'] = 'Uncorrelated'

    options['whiten'] = False

    options['kstep'] = kstep

    restrictions, blocks = SVAR.estPrepare.prepare_blocks_restrictions(n, n_rec, blocks, restrictions)
    options['restrictions'] = restrictions
    options['blocks'] = blocks

    moments = SVAR.estPrepare.prepare_moments('GMM', moments, addThirdMoments, addFourthMoments, moments_blocks,
                                              blocks, n,moments_MeanIndep)
    options['moments'] = moments
    options['moments_MeanIndep'] = moments_MeanIndep
    options['moments_powerindex'] = SVAR.SVARutilGMM.get_Moments_powerindex(moments)

    options['moments_num'] = np.shape(options['moments'])[0]
    options['moments_num2']  = np.sum(np.sum(options['moments'], axis=1) == 2)
    options['moments_num3']  = np.sum(np.sum(options['moments'], axis=1) == 3)
    options['moments_num4']  = np.sum(np.sum(options['moments'], axis=1) == 4)


    Jacobian = SVAR.SVARutilGMM.generate_Jacobian_function(moments=moments, restrictions=restrictions)
    options['Jacobian'] = Jacobian



    Sdel_func = SVAR.SVARutilGMM.generat_S_del_functions(moments, restrictions)
    options['Sdel_func'] = Sdel_func

    bstart = SVAR.estPrepare.prepare_bstart('GMM', bstart, u, options, bstartopt=bstartopt )
    options['bstart'] = bstart

    if S_func:
        if Wpara == 'Independent':
            S_func = SVAR.SVARutilGMM.generate_SIndep_function(moments)
        elif Wpara == 'Uncorrolated':
            S_func = False
        elif Wpara == 'Uncorrelated_uncentered':
            S_func = False
    options['S_func'] = S_func

    # Weighting
    if Wpara == 'Uncorrelated' or  Wpara == 'Independent'  or Wpara == 'Uncorrelated_uncentered':
        options['Wpara'] = Wpara
    else:
        print("Invalid Wpara value. Set Wpara=Uncorrelated")
        Wpara = 'Uncorrelated'
        options['Wpara'] = 'Uncorrelated'
    W = SVAR.estPrepare.prepare_W(u, n, W, Wstartopt, moments, bstart, restrictions,   Wpara, options['S_func']  )
    options['W'] = W



    if np.shape(options['moments'])[0] < np.shape(options['bstart'])[0] :
        raise ValueError('Less moment conditions than parameters. The SVAR is not identified')

    return options


def SVARout(est_SVAR, options, u):
    T, n = np.shape(u)
    out_SVAR = dict()
    out_SVAR['options'] = options

    b_est = est_SVAR['x']
    out_SVAR['b_est'] = b_est


    B_est = SVAR.get_BMatrix(b_est, restrictions=options['restrictions'], whiten=options['whiten'],
                             blocks=options['blocks'] )
    out_SVAR['B_est'] = B_est



    e = SVAR.innovation(u, b_est, restrictions=options['restrictions'], whiten=options['whiten'],
                        blocks=options['blocks'] )
    out_SVAR['e'] = e

    g = get_g(b=b_est, u=u, restrictions=options['restrictions'], moments=options['moments'],
              moments_powerindex=SVAR.SVARutilGMM.get_Moments_powerindex(options['moments']),
              whiten=options['whiten'], blocks=options['blocks'] )
    out_SVAR['g'] = g


    Omega_all = SVAR.SVARutil.get_Omega(e)
    out_SVAR['Omega_all'] = Omega_all
    omega = SVAR.SVARutil.get_Omega_Moments(e)
    CoOmega = SVAR.SVARutil.get_CoOmega(e)
    out_SVAR['omega'] = omega




    if options['Avarparametric'] == 'Independent':
        G = get_G_Indep(options['moments'], out_SVAR['B_est'], out_SVAR['omega'], options['restrictions'])
        S = get_S_Indep(options['moments'], options['moments'], omega=out_SVAR['omega'])
    elif options['Avarparametric'] == 'Uncorrelated':
        G = options['Jacobian'](u=u, b=out_SVAR['b_est'], restrictions=options['restrictions'], CoOmega=CoOmega)
        S = get_S(u, out_SVAR['b_est'],   options['moments'], options['restrictions'] )
    elif options['Avarparametric'] == 'Uncorrelated_uncentered':
        G = options['Jacobian'](u=u, b=out_SVAR['b_est'], restrictions=options['restrictions'], CoOmega=CoOmega)
        S = get_S_uncentered(u, out_SVAR['b_est'],   options['moments'], options['restrictions'] )



    if options['WupdateInOutput']:
        options['W'] = np.linalg.inv(S)

    V_est = get_Avar(n, G, S, W=options['W'], restrictions=options['restrictions'])
    out_SVAR['G'] = G
    out_SVAR['S'] = S
    out_SVAR['Avar_est'] = V_est


    loss = SVAR.estimatorGMM.loss(u, SVAR.get_BVector(out_SVAR['B_est'],  restrictions=options['restrictions']),
                                                      restrictions=options['restrictions'],
                                                      moments=options['moments'],
                                                      moments_powerindex=options['moments_powerindex'],
                                                      W=options['W'] )
    out_SVAR['loss'] = loss

    out_SVAR['J'] = np.multiply(T, loss)
    out_SVAR['Jpvalue'] = 1 - scipy.stats.chi2.cdf(out_SVAR['J'], np.size(g) - np.size(b_est))


    t_all = np.empty_like(b_est)
    t_all_p = np.empty_like(b_est)
    for idx, b in np.ndenumerate(b_est):
        avar = out_SVAR['Avar_est']
        avar = avar[np.logical_not(np.isnan(avar) ) ]
        avar = np.reshape(avar,[np.size(b_est),np.size(b_est)])
        avar_this = avar[idx,idx]

        t_this = T*np.divide(np.power(b_est[idx],2),avar_this)
        t_all[idx] = t_this
        t_all_p[idx] = 1 - scipy.stats.chi2.cdf(t_all[idx], 1)
    out_SVAR['t_all'] = t_all
    out_SVAR['t_all_p'] = t_all_p

    # Parameter wald
    out_SVAR['wald_all'], out_SVAR['wald_all_p'] = SVAR.SVARutilGMM.wald_param_all(B_est, options['restrictions'], out_SVAR['Avar_est'], T)

    # Fval - Rec
    out_SVAR['wald_rec'], out_SVAR['wald_rec_p']  = SVAR.SVARutilGMM.waldRec(B_est, out_SVAR['Avar_est'], options['restrictions'], T)

    if options['printOutput']:
            SVAR.estOutput.print_out(n, T, out_SVAR)


    return out_SVAR



import copy
import itertools
from types import FunctionType

import numpy as np

import SVAR
import scipy.stats

def get_Cr(r, n):
    # Generates indices of co-moment
    # Creat an array of all possible combinations of n numbers from 0 to n that sum up to r
    Cr_tmp = list(itertools.combinations_with_replacement(range(r), n))
    Cr = list()
    for cr in Cr_tmp:
        if sum(cr) == r:
            for this_cr in np.unique(list(itertools.permutations(cr)), axis=0):
                Cr.append(this_cr)
    Cr = np.asarray(Cr)

    # # improved code
    # Generates all possible combinations of n numbers from 0 to n that sum up to r
    # Cr_tmp = itertools.combinations_with_replacement(range(r), n)
    # Cr = []
    # used = set()
    # for cr in Cr_tmp:
    #     if sum(cr) == r:
    #         for this_cr in itertools.permutations(cr):
    #             if this_cr not in used:
    #                 Cr.append(this_cr)
    #                 used.add(this_cr)
    # Cr = np.asarray(Cr)
    return Cr






def get_Mr(r, n):
    # Generates indices of moments (e.g. variance)

    Mr = np.dot(np.eye(n), r)
    Mr = Mr.astype(int)
    return Mr

def get_Moments_MIcorrection(n, blocks=False):
    if blocks == False:
        blocks = list()
        blocks.append(np.array([1, n]))
    moments = np.full([1, n], 0)
    for block in blocks:
        n_min = block[0]
        n_max = block[1]
        block_length = 1 + n_max - n_min
        if block_length > 1:
            moments_this = get_Cr(4, block_length)
            select = np.full(np.shape(moments_this)[0], False)
            for idx, moment in enumerate(moments_this):
                if np.isin(2, moment) and not(np.isin(1, moment)):
                    select[idx] = True
            moments_this = moments_this[select]


            if np.shape(moments_this)[0] != 0:
                number_moments_this = np.shape(moments_this)[0]
                moments_this = np.hstack([np.zeros([number_moments_this, n_min - 1], dtype=int), moments_this])
                moments_this = np.hstack([moments_this, np.zeros([number_moments_this, n - n_max], dtype=int)])
                moments = np.append(moments, moments_this, axis=0)

    moments = moments[1:, :]
    return moments

def get_Moments(estimator, n, blocks=False, addThirdMoments=False, addFourthMoments=True,  moments_MeanIndep=False):
    if blocks == False:
        blocks = list()
        blocks.append(np.array([1, n]))
    moments = np.full([1, n], 0)

    if estimator == 'GMM' or estimator == 'CUE' or estimator == 'CSUE':
        moments = np.append(moments, get_Mr(2, n), axis=0)
        moments = np.append(moments, get_Cr(2, n), axis=0)

    for block in blocks:
        n_min = block[0]
        n_max = block[1]
        block_length = 1 + n_max - n_min
        if block_length > 1:
            moments_this = np.full([1, block_length], 0)
            if addThirdMoments:
                if estimator == 'GMM' or estimator == 'CUE' or estimator == 'CSUE':
                    moment_add = get_Cr(3, block_length)
                    moments_this = np.append(moments_this, moment_add, axis=0)

                elif estimator == 'GMM_W':
                    moment_add = get_Cr(3, block_length)

                    moments_this = np.append(moments_this, moment_add, axis=0)

                elif estimator == 'GMM_WF' or estimator == 'RidgeBWF':
                    moments_this = np.append(moments_this, get_Mr(3, block_length), axis=0)

            if addFourthMoments:
                if estimator == 'GMM' or estimator == 'CUE' or estimator == 'CSUE':
                    moment_add = get_Cr(4, block_length)

                    if moments_MeanIndep:
                        select = np.full(np.shape(moment_add)[0], False)
                        for idx, moment in enumerate(moment_add):
                            if np.isin(3, moment) or (np.isin(1, moment)):
                                select[idx] = True
                        moment_add = moment_add[select]

                    moments_this = np.append(moments_this, moment_add, axis=0)


                elif estimator == 'GMM_W':
                    moment_add = get_Cr(4, block_length)

                    if moments_MeanIndep:
                        select = np.full(np.shape(moment_add)[0], False)
                        for idx, moment in enumerate(moment_add):
                            if np.isin(3, moment) or (np.isin(1, moment)):
                                select[idx] = True
                        moment_add = moment_add[select]

                    moments_this = np.append(moments_this, moment_add, axis=0)

                elif estimator == 'GMM_WF' or estimator == 'RidgeBWF':
                    moments_this = np.append(moments_this, get_Mr(4, block_length), axis=0)
            moments_this = moments_this[1:, :]

            if np.shape(moments_this)[0] != 0:
                number_moments_this = np.shape(moments_this)[0]
                moments_this = np.hstack([np.zeros([number_moments_this, n_min - 1], dtype=int), moments_this])
                moments_this = np.hstack([moments_this, np.zeros([number_moments_this, n - n_max], dtype=int)])
                moments = np.append(moments, moments_this, axis=0)

    moments = moments[1:, :]
    return moments

def get_Moments_RecNonRed(n, addThirdMoments=False, addFourthMoments=True):
    moments = get_Moments('GMM', n, blocks=False, addThirdMoments=addThirdMoments, addFourthMoments=addFourthMoments)

    # unselect redundant moments
    select = np.full(np.shape(moments)[0], False)
    for idx, moment in enumerate(moments):
        if np.sum(moment) == 2:
                select[idx] = True
        elif np.sum(moment)==3:
            if np.isin(2,moment):
                select[idx] = True
        elif np.sum(moment)==4:
            if np.isin(3,moment):
                select[idx] = True
            elif not(np.isin(1,moment)):
                select[idx] = True

    moments = moments[select,:]



    return moments


def get_Moments_powerindex(moments):
    moments_fast = np.zeros([np.shape(moments)[0], np.shape(moments)[1] * 5], dtype=bool)
    counter = 0
    for moment in moments:
        for i in range(np.shape(moments)[1]):
            moments_fast[counter, i * 5 + moment[i]] = 1
        counter += 1
    return moments_fast



def get_Moment_transformed(moment):
    moment_trans = np.array([], dtype=int)
    for z in range(np.size(moment)):
        moment_trans = np.append(moment_trans, np.ones(moment[z], dtype=int) * (z + 1))
    return moment_trans



def calc_f(moments, moments_powerindex,e_power,T):
    # # Old code
    # counter = 0
    # f = np.empty([T, np.shape(moments)[0]])
    # for mom_fast in moments_powerindex:
    #     f[:, counter] = np.prod(e_power[:, mom_fast], axis=1)
    #     if ~np.isin(1, moments[counter, :]):
    #         if np.isin(2, moments[counter, :]):
    #             if np.sum(moments[counter, :]) == 2:
    #                 f[:, counter] = np.subtract(f[:, counter], 1)
    #             elif np.sum(moments[counter, :]) == 4:
    #                 f[:, counter] = np.subtract(f[:, counter], np.power(1,2))
    #         if np.isin(4, moments[counter, :]):
    #             f[:, counter] = np.subtract(f[:, counter], 3)
    #     counter += 1

    f = calc_f_improved(moments, moments_powerindex,e_power,T)
    return f

def calc_f_improved(moments, moments_powerindex,e_power,T):

    f = np.empty([T, np.shape(moments)[0]])
    # Iterate over the fast moments
    for i, mom_fast in enumerate(moments_powerindex):
        # Compute the powers of the exponents for the current moment
        powers = e_power[:, mom_fast]
        # Compute the products of the powers along axis 1
        f[:, i] = np.prod(powers, axis=1)


    k_array = np.zeros(moments.shape[0])
    k_array[~np.any(moments == 1, axis=1) & (np.sum(moments, axis=1)==2)] = 1
    k_array[~np.any(moments == 1, axis=1) & (np.sum(moments, axis=1)==4) & np.any(moments == 2, axis=1) ] = 1
    k_array[~np.any(moments == 1, axis=1) & np.any(moments == 4, axis=1)  ] = 3
    f = f - k_array
    return f



def get_f(u, b, restrictions, moments, moments_powerindex, whiten=False, blocks=False  ):
    T, n = np.shape(u)
    e = SVAR.innovation(u, b, restrictions=restrictions, whiten=whiten, blocks=blocks )
    e_power = np.empty([np.shape(e)[0], n * 5])
    e_power[:, range(0, 5 * n, 5)] = np.ones([np.shape(e)[0], n])
    for i in range(n):
        for j in range(1, 5):
            this_entry = (i) * 5 + j
            e_power[:, this_entry] = np.multiply(e_power[:, this_entry - 1], e[:, i])

    f = calc_f(moments, moments_powerindex,e_power,T)


    return f

def get_f_wf(u, b, restrictions, moments, moments_powerindex, whiten=True, blocks=False ):
    T, n = np.shape(u)
    e = SVAR.innovation(u, b, restrictions=restrictions, whiten=whiten, blocks=blocks )


    f = np.empty([T, np.shape(moments)[0]])
    # Iterate over the fast moments
    for i, mom in enumerate(moments):
        # Compute the powers of the exponents for the current moment
        powers = np.power(e,mom)
        # Compute the products of the powers along axis 1
        f[:, i] = np.prod(powers, axis=1)

    k_array = np.zeros(moments.shape[0])
    k_array[~np.any(moments == 1, axis=1) & (np.sum(moments, axis=1)==2)] = 1
    k_array[~np.any(moments == 1, axis=1) & (np.sum(moments, axis=1)==4) & np.any(moments == 2, axis=1) ] = 1
    k_array[~np.any(moments == 1, axis=1) & np.any(moments == 4, axis=1)  ] = 3
    f = f - k_array

    return f


def get_g(u, b, restrictions, moments, moments_powerindex, whiten=False, blocks=False  ):
    return np.mean(
        get_f(b=b[:], u=u[:], restrictions=restrictions[:], moments=moments[:], moments_powerindex=moments_powerindex,
              whiten=whiten, blocks=blocks  ), axis=0)

def get_g_wf(u, b, restrictions, moments, moments_powerindex, whiten=True, blocks=False  ):
    return np.mean(
        get_f_wf(b=b[:], u=u[:], restrictions=restrictions[:], moments=moments[:], moments_powerindex=moments_powerindex,
              whiten=whiten, blocks=blocks ), axis=0)

def generat_S_del_ql_function(moments,q,l):
    n = np.shape(moments)[1]

    function_string = str("def f(omegaext,A): ")
    function_string = function_string + str('S=np.array([')
    for countx, moment_x in enumerate(moments):
        if countx != 0:
            function_string = function_string + str(',')

        this_moment_string = str('[ ')
        for county, moment_y in enumerate(moments):
            if county != 0:
                this_moment_string = this_moment_string + str(',')

            moment_this = moment_x + moment_y

            omega_str = SVAR.SVARutilGMM.get_omegaextstring(moment_this, n, q, l)

            if ~np.isin(1, moment_x):
                omega_str2 = SVAR.SVARutilGMM.get_omegaextstring(moment_y, n, q, l)
                omega_str = omega_str + '-' + omega_str2
            if ~np.isin(1, moment_y):
                omega_str2 = SVAR.SVARutilGMM.get_omegaextstring(moment_x, n, q, l)
                omega_str = omega_str + '-' + omega_str2
            if ~np.isin(1, moment_y) and ~np.isin(1, moment_x):
                omega_str = omega_str + '+ 1'

            this_moment_string = this_moment_string + omega_str
        this_moment_string = this_moment_string + str(']')
        function_string = function_string + this_moment_string
    function_string = function_string + str(']);')
    function_string = function_string + str('return S')

    function_code = compile(function_string, "<string>", "exec")
    function_func = FunctionType(function_code.co_consts[0], globals(), "f")

    return function_func


def generat_S_del_functions(moments,restrictions):
    n = np.shape(moments)[1]
    Sdelbql_func_all = list()
    for q in range(n):
        for l in range(n):
            # only if B element is not restricted
            if np.isnan(restrictions[q, l]):
                Sdelbql_func_all.append(SVAR.SVARutilGMM.generat_S_del_ql_function(moments, q, l))

    return Sdelbql_func_all

def generate_Jacobian_function(moments, restrictions):
    n = np.shape(moments)[1]
    function_string = str("def f(u,b,restrictions,CoOmega): ")
    function_string = function_string + str('B = SVAR.get_BMatrix(b, restrictions=restrictions);')
    function_string = function_string + str('A = np.linalg.inv(B);')

    function_string = function_string + str('[CoOmega1, CoOmega2, CoOmega3, CoOmega4, CoOmega5, CoOmega6, CoOmega7, CoOmega8] = CoOmega;')

    function_string = function_string + str('G=np.array([')
    counter = 0
    for moment in moments:
        counter_inner = 0
        if counter != 0:
            function_string = function_string + str(',')
        # Calculate Jacboian of this moment
        this_moment_string = str('[ ')
        for q in range(0, n):
            for l in range(0, n):
                # only if B element is not restricted
                if np.isnan(restrictions[q, l]):
                    tmp_string = momentderiv(moment,q,l,n)

                    if counter_inner != 0:
                        this_moment_string = this_moment_string + str(',')
                    counter_inner += 1
                    this_moment_string = this_moment_string + tmp_string
        this_moment_string = this_moment_string + str(']')

        # Append to function_string
        function_string = function_string + this_moment_string
        counter += 1

    function_string = function_string + str(']);')
    function_string = function_string + str('return G')
    # print(function_string)

    function_code = compile(function_string, "<string>", "exec")
    function_func = FunctionType(function_code.co_consts[0], globals(), "f")

    return function_func

def generate_SIndep_function(moments):
    function_string = str("def f(omega): ")
    function_string = function_string + str('S=np.array([')
    for countx, moment_x in enumerate(moments):
        if countx != 0:
            function_string = function_string + str(',')

        this_moment_string = str('[ ')
        for county, moment_y in enumerate(moments):
            if county != 0:
                this_moment_string = this_moment_string + str(',')

            # symmetric matrix
            if countx>county:
                omega_str = '0'
            else:
                moment_this = moment_x + moment_y

                omega_str = SVAR.SVARutilGMM.get_omegastring(moment_this)

                if ~np.isin(1, moment_x):
                    omega_str2 = SVAR.SVARutilGMM.get_omegastring(moment_y)
                    omega_str = omega_str + '-' + omega_str2
                if ~np.isin(1, moment_y):
                    omega_str2 = SVAR.SVARutilGMM.get_omegastring(moment_x)
                    omega_str = omega_str + '-' + omega_str2
                if ~np.isin(1, moment_y) and ~np.isin(1, moment_x):
                    omega_str = omega_str + '+ 1'

            this_moment_string = this_moment_string + omega_str
        this_moment_string = this_moment_string + str(']')
        function_string = function_string + this_moment_string
    function_string = function_string + str(']);')
    function_string = function_string + str('return S')
    # print(function_string)
    function_code = compile(function_string, "<string>", "exec")
    function_func = FunctionType(function_code.co_consts[0], globals(), "f")

    return function_func


def eval_CoOmega(moment,CoOmega):
    [CoOmega1, CoOmega2, CoOmega3, CoOmega4, CoOmega5, CoOmega6, CoOmega7, CoOmega8] = CoOmega

    moment_trans = get_Moment_transformed(moment)
    if sum(moment) == 0:
        omega = 1
    elif sum(moment) == 1:
        omega = CoOmega1[moment_trans[0]  - 1     ]
    elif sum(moment) == 2:
        omega = CoOmega2[moment_trans[0]  - 1  , moment_trans[1]  - 1  ]
    elif sum(moment) == 3:
        omega = CoOmega3[moment_trans[0]  - 1  , moment_trans[1]  - 1, moment_trans[2]  - 1   ]
    elif sum(moment) == 4:
        omega = CoOmega4[moment_trans[0]  - 1  , moment_trans[1]  - 1, moment_trans[2]  - 1, moment_trans[3]  - 1    ]
    elif sum(moment) == 5:
        omega = CoOmega5[moment_trans[0] - 1, moment_trans[1] - 1, moment_trans[2] - 1, moment_trans[3] - 1, moment_trans[4] - 1]
    elif sum(moment) == 6:
        omega = CoOmega6[moment_trans[0] - 1, moment_trans[1] - 1, moment_trans[2] - 1, moment_trans[3] - 1, moment_trans[4] - 1, moment_trans[5] - 1]
    elif sum(moment) == 7:
        omega = CoOmega7[moment_trans[0] - 1, moment_trans[1] - 1, moment_trans[2] - 1, moment_trans[3] - 1, moment_trans[4] - 1, moment_trans[5] - 1, moment_trans[6] - 1]
    elif sum(moment) == 8:
        omega = CoOmega8[moment_trans[0] - 1, moment_trans[1] - 1, moment_trans[2] - 1, moment_trans[3] - 1, moment_trans[4] - 1, moment_trans[5] - 1, moment_trans[6] - 1, moment_trans[7] - 1]

    else:
        print("error")

    return omega

def get_CoOmegaEvalString(moment):
    moment_trans = get_Moment_transformed(moment)
    if sum(moment) == 0:
        omega_str = '1'
    elif sum(moment) == 1:
        omega_str = 'CoOmega1[' + str(int(moment_trans[0]) - 1) +   ']'
    elif sum(moment) == 2:
        omega_str = 'CoOmega2[' + str(int(moment_trans[0]) - 1) + ', ' + str(
            int(moment_trans[1]) - 1) + ']'
    elif sum(moment) == 3:
        omega_str = 'CoOmega3[' + str(int(moment_trans[0]) - 1) + ', ' + str(
            int(moment_trans[1]) - 1) + ', ' + str(int(moment_trans[2]) - 1) + ']'
    elif sum(moment) == 4:
        omega_str = 'CoOmega4[' + str(int(moment_trans[0]) - 1) + ', ' + str(
            int(moment_trans[1]) - 1) + ', ' + str(
            int(moment_trans[2]) - 1) + ', ' + str(int(moment_trans[3]) - 1) + ']'
    elif sum(moment) == 5:
        omega_str = 'CoOmega5[' \
                    + str(int(moment_trans[0]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[1]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[2]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[3]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[4]) - 1) + ']'
    elif sum(moment) == 6:
        omega_str = 'CoOmega6[' \
                    + str(int(moment_trans[0]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[1]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[2]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[3]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[4]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[5]) - 1)  + ']'
    elif sum(moment) == 7:
        omega_str = 'CoOmega7[' \
                    + str(int(moment_trans[0]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[1]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[2]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[3]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[4]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[5]) - 1)\
                    + ', ' \
                    + str(int(moment_trans[6]) - 1)   + ']'
    elif sum(moment) == 8:
        omega_str = 'CoOmega8[' \
                    + str(int(moment_trans[0]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[1]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[2]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[3]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[4]) - 1) \
                    + ', ' \
                    + str(int(moment_trans[5]) - 1)\
                    + ', ' \
                    + str(int(moment_trans[6]) - 1)\
                    + ', ' \
                    + str(int(moment_trans[7]) - 1)    + ']'
    else:
        omega_str = '999'
    return omega_str


def get_omegastring(moment):
    omega_str = ''
    if sum(moment) == 0:
        omega_str = '1'
    else:
        for countmom, mom in enumerate(moment):
            if mom > 0:
                if not (omega_str == ''):
                    omega_str = omega_str + '*'
                omega_str = omega_str + 'omega[' + str(int(countmom)) + ', ' + str(int(mom) - 1) + ']'

    return omega_str

def get_omegaextstring(moment,n,q,l):
    omega_str = ''
    for k, mom in enumerate(moment):
        if k > 0:
            omega_str = omega_str + '+'
        for idx in range(n):
            if idx > 0:
                omega_str = omega_str + '*'

            if idx == k:
                if moment[idx] == 0:
                    omega_str = omega_str + str(0)
                else:
                    omega_str = omega_str + str(int(mom)) +'*'+ 'A[' + str(k) + ',' + str(q) +']*' + 'omegaext[' + str(int(idx)) + ', ' + str(int(moment[idx]) - 1) + ', ' + str(l+1) + ']'
            else:
                omega_str = omega_str + 'omegaext[' + str(int(idx)) + ', ' + str(int(moment[idx])  ) + ', 0'  + ']'


    return omega_str


def momentderiv(moment,q,l,n):
    tmp_string = str(' ')

    for k in range(0, n):
        if moment[k] != 0:
            moment_new = copy.deepcopy(moment)
            moment_new[l] = moment_new[l] + 1
            moment_new[k] = moment_new[k] - 1

            omega_str = get_CoOmegaEvalString(moment_new)

            if moment[k] != 0:
                tmp_string = tmp_string + str(
                    '-A[' + str(k) + ',' + str(q) + '] * ' + str(moment[k]) + ' * ' + omega_str)
    return tmp_string

def get_G_Indep(Moments, B, omega, B_restrictions=[]):
    # Calculates analytical G = E[ partial f(u,B_0) / partial B ]

    n, n = np.shape(B)

    if np.array(B_restrictions).size == 0: B_restrictions = np.full([n, n], np.nan)

    A = np.linalg.inv(B)
    n = np.shape(B)[1]
    # add one row to omega for moment e^0=1
    omega = np.concatenate((np.ones([n, 1]), omega), axis=1)
    # empty G array
    G = np.empty([np.shape(Moments)[0], np.sum(np.isnan(B_restrictions))])
    moment_counter = 0
    for moment in Moments:
        elementcounter = 0
        for i in range(0, n):
            for j in range(0, n):
                # only if B element is not restricted
                if np.isnan(B_restrictions[i, j]):
                    G_this = 0
                    for idx in range(0, n):
                        moment_new = copy.deepcopy(moment)
                        moment_new[j] = moment_new[j] + 1
                        moment_new[idx] = moment_new[idx] - 1
                        G_this = G_this - A[idx, i] * moment[idx] * np.prod(omega[np.array([range(0, n)]), moment_new])
                    G[moment_counter, elementcounter] = G_this
                    elementcounter = elementcounter + 1
        moment_counter = moment_counter + 1
    return G



def get_G(Moments, B,  CoOmega, B_restrictions=[]):
    # Calculates analytical G = E[ partial f(u,B_0) / partial B ]

    n, n = np.shape(B)

    if np.array(B_restrictions).size == 0: B_restrictions = np.full([n, n], np.nan)

    A = np.linalg.inv(B)
    n = np.shape(B)[1]
    # empty G array
    G = np.empty([np.shape(Moments)[0], np.sum(np.isnan(B_restrictions))])
    moment_counter = 0
    for moment in Moments:
        elementcounter = 0
        for i in range(0, n):
            for j in range(0, n):
                # only if B element is not restricted
                if np.isnan(B_restrictions[i, j]):
                    G_this = 0
                    for idx in range(0, n):
                        moment_new = copy.deepcopy(moment)
                        moment_new[j] = moment_new[j] + 1
                        moment_new[idx] = moment_new[idx] - 1

                        # catch negative values. Unimportant derivative is zero
                        if moment_new[idx] < 0:
                            moment_new[idx] = 0

                        G_this = G_this - A[idx, i] * moment[idx] * eval_CoOmega(moment_new, CoOmega)

                    G[moment_counter, elementcounter] = G_this
                    elementcounter = elementcounter + 1
        moment_counter = moment_counter + 1
    return G

def get_S_Indep(Moments_1, Moments_2, omega):
    # Calculates analytically S = E[ f_[Moments_1](u,B_0) f_[Moments_2](u,B_0)' ]

    # empty S array
    S = np.empty([np.shape(Moments_1)[0], np.shape(Moments_2)[0]])
    n = np.shape(Moments_1)[1]
    # add one row to omega for moment e^0=1
    omega = np.concatenate((np.ones([n, 1]), omega), axis=1)
    moments_1_counter = 0
    moments_2_counter = 0
    for moment_1 in Moments_1:
        for moment_2 in Moments_2:
            v_1 = - int(~np.isin(1, moment_1))
            v_2 = - int(~np.isin(1, moment_2))
            S[moments_1_counter, moments_2_counter] = np.prod(
                omega[np.array([range(0, n)]), moment_1 + moment_2]) + v_1 * np.prod(
                omega[np.array([range(0, n)]), moment_2]) + v_2 * np.prod(
                omega[np.array([range(0, n)]), moment_1]) + v_1 * v_2
            moments_2_counter = moments_2_counter + 1
        moments_2_counter = 0
        moments_1_counter = moments_1_counter + 1
    return S


def get_S(u, b,   moments, restrictions ):
    moments_powerindex = get_Moments_powerindex(moments)
    f = get_f(u=u, b=b, restrictions=restrictions, moments=moments,
              moments_powerindex=moments_powerindex)
    S = np.cov(f.T )
    return S

def get_S_uncentered(u, b,   moments, restrictions ):
    [T,n] = np.shape(u)
    moments_powerindex = get_Moments_powerindex(moments)
    f = get_f(u=u, b=b, restrictions=restrictions, moments=moments,
              moments_powerindex=moments_powerindex)
    S =  np.dot(f.T, f) / (f.shape[0] - 1)
    return S


def wald(R,r,b,avar,T):
    wald2 = avar

    wald1 = np.dot(R, b) - r
    wald3 = np.linalg.multi_dot([R, wald2, np.transpose(R)])
    try:
        wald3 = np.linalg.inv(wald3)
        wald = np.multiply(T, np.linalg.multi_dot([np.transpose(wald1), wald3, wald1]))

    except:
        wald3 = np.divide(1, wald3)
        wald = T * wald1 * wald3 * wald1

    p = 1 - scipy.stats.chi2.cdf(wald, np.shape(R)[0])

    return wald,p

def wald_param_all(B,restrictions,avar,T):
    b = SVAR.get_BVector(B, restrictions=restrictions, whiten=False)
    wald_all = np.empty_like(b)
    wald_all_p = np.empty_like(b)
    avar = avar[np.logical_not(np.isnan(avar))]
    avar = np.reshape(avar, [np.size(b), np.size(b)])
    for idx, b_this in np.ndenumerate(b):
        R = np.full([1, np.size(b)], 0)
        R[0, idx] = 1
        r = 0
        wald_all[idx], wald_all_p[idx] = SVAR.SVARutilGMM.wald(R, r, b, avar, T)
    return wald_all, wald_all_p

def waldRec(B,avar,restrictions,T):
    restrictions_rec = SVAR.SVARutil.getRestrictions_recursive(B)
    b =  SVAR.get_BVector(B, restrictions=restrictions, whiten=False)
    select = (np.isnan(SVAR.get_BVector(restrictions_rec))) == False
    select = select[(np.isnan(SVAR.get_BVector(restrictions)))]
    counter = 0
    R = np.full([np.sum(select), np.size(b)], 0)
    for idx, b_this in np.ndenumerate(b):
        if select[idx]:
            R[counter, idx] = 1
            counter += 1
    r = np.full([np.sum(select)], 0)
    avar = avar[np.logical_not(np.isnan(avar))]
    avar = np.reshape(avar, [np.size(b), np.size(b)])
    wald_rec, wald_rec_p = SVAR.SVARutilGMM.wald(R, r, b, avar, T)
    return wald_rec, wald_rec_p

def waldRest(B,avar,restrictions,T,testrest):
    b = SVAR.get_BVector(B, restrictions=restrictions, whiten=False)
    select = (np.isnan(SVAR.get_BVector(testrest))) == False
    select = select[(np.isnan(SVAR.get_BVector(restrictions)))]
    counter = 0
    R = np.full([np.sum(select), np.size(b)], 0)
    for idx, b_this in np.ndenumerate(b):
        if select[idx]:
            R[counter, idx] = 1
            counter += 1
    r = np.full([np.sum(select)], SVAR.get_BVector(testrest[np.isnan(restrictions)])[select])
    avar = avar[np.logical_not(np.isnan(avar))]
    avar = np.reshape(avar, [np.size(b), np.size(b)])
    wald_stat, wald_stat_p = SVAR.SVARutilGMM.wald(R, r, b, avar, T)
    return wald_stat, wald_stat_p

def LRRest(u,SVAR_outsave,testrest):
    options = copy.deepcopy(SVAR_outsave['options'])
    restrictions = options['restrictions']
    restrictions[np.isnan(restrictions)] = testrest[np.isnan(restrictions)]
    options['restrictions'] = restrictions
    options['bstart'] = options['bstart'][np.isnan(SVAR.get_BVector(restrictions))]
    Jacobian = SVAR.SVARutilGMM.generate_Jacobian_function(moments=options['moments'], restrictions=restrictions)
    options['Jacobian'] = Jacobian

    SVAR_outrest = SVAR.SVARest(u, estimator=SVAR_outsave['options']['estimator'], options=options,prepared=True)
    lr_stat = options['T'] * (SVAR_outrest['loss'] - SVAR_outsave['loss'])
    lr_stat_p = 1 - scipy.stats.chi2.cdf(lr_stat, np.sum((np.isnan(SVAR.get_BVector(testrest))) == False))
    return lr_stat, lr_stat_p


def LMRest(u, SVAR_outsave, testrest):
    options = copy.deepcopy(SVAR_outsave['options'])
    restrictions = options['restrictions']
    restrictions[np.isnan(restrictions)] = testrest[np.isnan(restrictions)]
    options['restrictions'] = restrictions
    options['bstart'] = options['bstart'][np.isnan(SVAR.get_BVector(restrictions))]
    Jacobian = SVAR.SVARutilGMM.generate_Jacobian_function(moments=options['moments'], restrictions=restrictions)
    options['Jacobian'] = Jacobian

    SVAR_outrest = SVAR.SVARest(u, estimator=SVAR_outsave['options']['estimator'], options=options, prepared=True)

    options_unrest = SVAR_outsave['options']
    options = SVAR_outrest['options']
    CoOmega = SVAR.SVARutil.get_CoOmega(SVAR_outrest['e'])

    g = copy.deepcopy(SVAR_outsave['g']) # ToDo: Wrong? g at Brest?
    G = copy.deepcopy(SVAR_outsave['G'])
    S = copy.deepcopy(SVAR_outsave['S'])

    lm_stat1 = np.matmul(np.matmul(np.transpose(g), np.linalg.inv(S)), G)
    lm_stat2 = np.linalg.inv(np.matmul(np.matmul(np.transpose(G), np.linalg.inv(S)), G))
    lm_stat3 = np.matmul(np.matmul(np.transpose(G), np.linalg.inv(S)), g)
    lm_stat = options['T'] * np.matmul(np.matmul(lm_stat1, lm_stat2), lm_stat3)
    lm_stat_p = 1 - scipy.stats.chi2.cdf(lm_stat, np.sum((np.isnan(SVAR.get_BVector(testrest))) == False))
    return lm_stat, lm_stat_p

def get_W_opt(u, b, restrictions, moments,     Wpara='Uncorrelated',  S_func=False  ):
    if Wpara=='Uncorrelated':
        S = SVAR.SVARutilGMM.get_S(u, b=b, moments=moments,restrictions=restrictions)
    elif Wpara=='Uncorrelated_uncentered':
        S = SVAR.SVARutilGMM.get_S_uncentered(u, b=b, moments=moments,restrictions=restrictions)
    elif Wpara == 'Independent':
        e = SVAR.innovation(u,b,restrictions=restrictions)
        omega = SVAR.SVARutil.get_Omega_Moments(e)
        if S_func==False:
            S = SVAR.SVARutilGMM.get_S_Indep(Moments_1=moments, Moments_2=moments, omega=omega )
        else:
            S = S_func(omega)
            S = S + S.T - np.diag(np.diag(S))

    try:
        W = np.linalg.inv(S)
    except:
        W = np.linalg.pinv(S)

    return W



import os
import numpy as np

import SVAR.SVARutilGMM
import SVAR.estSVAR
import SVAR.SVARbasics
import pickle
import pytest
import SVAR.estimatorGMM
import SVAR.estimatorGMMW
import SVAR.SVARutil

np.random.seed(0)

if False:
    with open("tests\data\eps.data", 'rb') as filehandle:
        # read the data as binary data stream
        eps = pickle.load(filehandle)
    supply_eps = eps


#
# # path = os.path.join("MCResults", version)
# file_name = "W.data"
# with open(file_name, 'wb') as filehandle:
#     pickle.dump(grad, filehandle)


@pytest.fixture
def supply_eps():
    with open("data/eps.data", 'rb') as filehandle:
        # read the data as binary data stream
        eps = pickle.load(filehandle)
    return eps




def test_GMM_avar(supply_eps):
    n = 2
    u = supply_eps[:, :n]
    restrictions = np.full([n, n], np.nan)
    B = np.eye(n)
    b = SVAR.SVARutil.get_BVector(B, restrictions=restrictions)
    moments = SVAR.SVARutilGMM.get_Mr(2, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(3, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(4, n), axis=0)
    W = np.eye(np.shape(moments)[0])
    k_step = 1
    # GMM_out = SVAR.K_GMM.GMM(u, b, Jacobian, W, restrictions, moments, kstep=k_step)
    opt_svar = dict()
    opt_svar['W'] = W
    opt_svar['Wstartopt'] = 'specific'
    opt_svar['kstep'] = k_step
    opt_svar['bstartopt'] = 'specific'
    opt_svar['bstart'] = b
    opt_svar['moments'] = moments
    opt_svar['restrictions'] = restrictions
    opt_svar['Avarparametric'] = "Independent"
    opt_svar['WupdateInOutput'] = False


    svar_out = SVAR.SVARest(u, estimator='GMM', prepOptions=opt_svar)
    V = svar_out['Avar_est']
    V_expected = np.array([[1.9402818468659488, 0.2733512070290248, 0.1284802727181451, 0.7012494706027909], [0.2733512070290248, 1.9756584345286183, -0.8135648505469544, 0.08800376364855292], [0.12848027271814508, -0.8135648505469546, 2.0956674150027013, 0.31668421313059375], [0.7012494706027907, 0.08800376364855292, 0.31668421313059375, 1.968376039877903]])

    V = np.round(V, 5)
    V_expected = np.round(V_expected, 5)

    assert V.tolist()  == V_expected.tolist()

    opt_svar['Avarparametric'] = "Uncorrelated"
    svar_out = SVAR.SVARest(u, estimator='GMM', prepOptions=opt_svar)
    V = svar_out['Avar_est']
    V_expected = np.array([[1.7409070057869636, -0.0317006809061288, 0.15592176107093303, 0.46321910091505997], [-0.03170068090612883, 1.4952798978427135, -0.6335438399532474, -0.3004910762201573], [0.155921761070933, -0.6335438399532474, 1.7048115431668374, 0.3879469041436635], [0.46321910091505986, -0.3004910762201573, 0.3879469041436635, 1.6913747391778382]])

    V = np.round(V, 5)
    V_expected = np.round(V_expected, 5)

    assert V.tolist() == V_expected.tolist()



def test_GMM(supply_eps):
    n = 2
    u = supply_eps[:, :n]
    restrictions = np.full([n, n], np.nan)
    B = np.eye(n)
    b = SVAR.SVARutil.get_BVector(B, restrictions=restrictions)
    moments = SVAR.SVARutilGMM.get_Mr(2, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(3, n), axis=0)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(4, n), axis=0)
    W = np.eye(np.shape(moments)[0])
    k_step = 1
    # GMM_out = SVAR.K_GMM.GMM(u, b, Jacobian, W, restrictions, moments, kstep=k_step)
    prepOptions = dict()
    prepOptions['W'] = W
    prepOptions['Wstartopt'] = 'specific'
    prepOptions['kstep'] = k_step
    prepOptions['bstartopt'] = 'specific'
    prepOptions['bstart'] = b
    prepOptions['moments'] = moments
    prepOptions['restrictions'] = restrictions
    prepOptions['Avarparametric'] = "Independent"
    svar_out = SVAR.SVARest(u, estimator='GMM', prepOptions=prepOptions)
    B_est = svar_out['B_est']
    loss = svar_out['loss']
    avar = svar_out['Avar_est']

    avar_expected = np.array([[1.9402818468659488, 0.2733512070290248, 0.1284802727181451, 0.7012494706027909], [0.2733512070290248, 1.9756584345286183, -0.8135648505469544, 0.08800376364855292], [0.12848027271814508, -0.8135648505469546, 2.0956674150027013, 0.31668421313059375], [0.7012494706027907, 0.08800376364855292, 0.31668421313059375, 1.968376039877903]])
    loss_expected = 0.0010397156643544272
    B_est_expected = np.array([[0.9935424993853278, -0.008754219586363085], [0.010958149691157675, 1.0134172158135832]])

    loss = np.round(loss, 5)
    loss_expected = np.round(loss_expected, 5)
    B_est = np.round(B_est, 5)
    B_est_expected = np.round(B_est_expected, 5)
    avar = np.round(avar, 5)
    avar_expected = np.round(avar_expected, 5)

    assert loss == loss_expected
    assert svar_out['options']['bstart'].tolist() == b.tolist()
    assert svar_out['options']['moments'].tolist() == moments.tolist()
    assert svar_out['options']['kstep'] == k_step
    assert B_est.tolist() == B_est_expected.tolist()
    assert svar_out['options']['W'].tolist() == W.tolist()
    assert avar.tolist() == avar_expected.tolist()

    prepOptions = dict()
    prepOptions['Wstartopt'] = 'I'
    svar_out = SVAR.SVARest(u, estimator='GMM', prepOptions=prepOptions)
    B_est = svar_out['B_est']
    loss = svar_out['loss']
    loss_expected = 0.0005985534575471171
    B_est_expected = np.array([[0.99407, -0.00511], [0.01775, 1.01315]])

    loss = np.round(loss, 5)
    loss_expected = np.round(loss_expected, 5)
    B_est = np.round(B_est, 5)
    B_est_expected = np.round(B_est_expected, 5)

    assert loss == loss_expected
    assert B_est.tolist() == B_est_expected.tolist()

    prepOptions = dict()
    prepOptions['bstartopt'] = 'GMM_WF'
    prepOptions['Wstartopt'] = 'WoptBstart'
    prepOptions['kstep'] = 1
    svar_out = SVAR.SVARest(u, estimator='GMM', prepOptions=prepOptions)
    B_est = svar_out['B_est']
    loss = svar_out['loss']
    loss_expected = 0.0006255099149225711
    B_est_expected = np.array([[0.99367, -0.0051], [0.01749, 1.01291]])

    loss = np.round(loss, 5)
    loss_expected = np.round(loss_expected, 5)
    B_est = np.round(B_est, 5)
    B_est_expected = np.round(B_est_expected, 5)


    assert loss == loss_expected
    assert B_est.tolist() == B_est_expected.tolist()

    prepOptions = dict()
    prepOptions['bstartopt'] = 'I'
    prepOptions['Wstartopt'] = 'WoptBstart'
    prepOptions['kstep'] = 1
    svar_out = SVAR.SVARest(u, estimator='GMM', prepOptions=prepOptions)
    B_est = svar_out['B_est']
    loss = svar_out['loss']
    loss_expected = 0.0005942712186022502
    B_est_expected = np.array([[0.9944162541003781, -0.004656693154018139], [0.01728797491300107, 1.0129375965423226]])

    loss = np.round(loss, 5)
    loss_expected = np.round(loss_expected, 5)
    B_est = np.round(B_est, 5)
    B_est_expected = np.round(B_est_expected, 5)

    assert loss == loss_expected
    assert B_est.tolist() == B_est_expected.tolist()


def test_GMM_white(supply_eps):
    n = 2
    u = supply_eps[:, :n]
    b = np.array([0])
    moments = SVAR.SVARutilGMM.get_Cr(3, n)
    moments = np.append(moments, SVAR.SVARutilGMM.get_Cr(4, n), axis=0)
    W = np.eye(np.shape(moments)[0])

    prepOptions = dict()
    prepOptions['W'] = W
    prepOptions['Wstartopt'] = 'specific'
    prepOptions['bstart'] = b
    prepOptions['bstartopt'] = 'specific'
    prepOptions['moments'] = moments
    svar_out = SVAR.SVARest(u, estimator='GMM_W', prepOptions=prepOptions)
    B_est = svar_out['B_est']
    loss = svar_out['loss']
    Omega2 = svar_out['Omega_all'][0]
    Omega2 = np.round(Omega2, 15)


    loss_expected = 0.003865450152482355
    B_est_expected = np.array([[0.997230283623227, -0.004399373906083158], [0.014447338782992897, 1.017457148695816]])
    Omega2_expected = np.array([[1.0, -0.0], [np.nan, 1.0]])

    loss = np.round(loss, 5)
    loss_expected = np.round(loss_expected, 5)
    B_est = np.round(B_est, 5)
    B_est_expected = np.round(B_est_expected, 5)
    Omega2 = np.round(Omega2, 5)
    Omega2_expected = np.round(Omega2_expected, 5)

    assert loss == loss_expected
    assert B_est.tolist() == B_est_expected.tolist()
    assert Omega2[~np.isnan(Omega2)].tolist() == Omega2_expected[~np.isnan(Omega2)].tolist()

def test_GMM_fast(supply_eps):
    n = 2
    u = supply_eps[:, :n]
    b = np.array([0])

    prepOptions = dict()
    prepOptions['bstart'] = b
    prepOptions['bstartopt'] = 'specific'
    svar_out = SVAR.SVARest(u, estimator='GMM_WF', prepOptions=prepOptions)
    B_est = svar_out['B_est']
    loss = svar_out['loss']
    Omega2 = svar_out['Omega_all'][0]
    Omega2 = np.round(Omega2, 15)

    loss_expected = -19.7494828332592
    B_est_expected = np.array([[0.997231567472602, -0.004098036379113462], [0.014139888759238346, 1.0174614678581229]])
    Omega2_expected = np.array([[1.0, -0.0], [np.nan, 1.0]])

    loss = np.round(loss, 5)
    loss_expected = np.round(loss_expected, 5)
    B_est = np.round(B_est, 5)
    B_est_expected = np.round(B_est_expected, 5)
    Omega2 = np.round(Omega2, 5)
    Omega2_expected = np.round(Omega2_expected, 5)


    assert loss == loss_expected
    assert B_est.tolist() == B_est_expected.tolist()
    assert Omega2[~np.isnan(Omega2)].tolist() == Omega2_expected[~np.isnan(Omega2)].tolist()

def test_PML(supply_eps):
    n = 2
    u = supply_eps[:, :n]
    b = np.array([0])

    prepOptions = dict()
    prepOptions['bstart'] = b
    prepOptions['bstartopt'] = 'specific'
    svar_out = SVAR.SVARest(u, estimator='PML', prepOptions=prepOptions)
    B_est = svar_out['B_est']
    loss = svar_out['loss']
    Omega2 = svar_out['Omega_all'][0]

    loss_expected = 4257.468466265423
    B_est_expected = np.array([[0.9972396384007051, 0.0008346666749092252], [0.009106998762809449, 1.0175189617907199]])
    Omega2_expected = np.array([[ 1., -0.], [np.nan,  1.]])

    loss = np.round(loss, 5)
    loss_expected = np.round(loss_expected, 5)
    B_est = np.round(B_est, 5)
    B_est_expected = np.round(B_est_expected, 5)
    Omega2 = np.round(Omega2, 5)
    Omega2_expected = np.round(Omega2_expected, 5)


    assert loss == loss_expected
    assert B_est.tolist() == B_est_expected.tolist()
    assert Omega2[~np.isnan(Omega2)].tolist() == Omega2_expected[~np.isnan(Omega2)].tolist()


def test_GMM_PartlyRecurisve(supply_eps):
    n = 5
    u = supply_eps[:, :n]

    n_rec = 2
    prepOptions = dict()
    prepOptions['n_rec'] = n_rec
    prepOptions['Wstartopt'] = 'I'
    SVAR_out = SVAR.SVARest(u, estimator='GMM_W', prepOptions=prepOptions)
    B_est_PC = SVAR_out['B_est']

    prepOptions = dict()
    prepOptions['n_rec'] = 5
    prepOptions['Wstartopt'] = 'I'
    Rec_out = SVAR.SVARest(u, estimator='GMM', prepOptions=prepOptions)
    B_est_rec = Rec_out['B_est']

    B_est_rec = np.round(B_est_rec, 5)
    B_est_PC = np.round(B_est_PC, 5)

    assert B_est_PC[:n_rec,:n_rec].tolist() == B_est_rec[:n_rec,:n_rec].tolist()

    expected = np.array([[0.003330608469677456, -0.033621407028708195],
 [-0.010995613002447795, -0.020220644924939964],
 [-0.0179660316466463, -0.0012376406712120119]])

    B_est_PC = np.round(B_est_PC, 5)
    expected = np.round(expected, 5)

    assert B_est_PC[n_rec:,:n_rec].tolist() == expected.tolist()

    expected = np.array([[0.9845972613833915, 0.02681168147749875, 0.004427083644211701],
 [-0.003259823923889076, 1.001091873682849, 0.00533142301218077],
 [0.0022816522163611945, 0.011374323197186675, 0.988404250711615]])

    B_est_PC = np.round(B_est_PC, 5)
    expected = np.round(expected, 5)

    assert B_est_PC[n_rec:, n_rec:].tolist() == expected.tolist()

    n_rec = 2
    prepOptions = dict()
    prepOptions['n_rec'] = n_rec
    prepOptions['bstartopt'] = 'GMM_WF'
    prepOptions['Wstartopt'] = 'WoptBstart'
    prepOptions['kstep'] = 1
    SVAR_out = SVAR.SVARest(u, estimator='GMM', prepOptions=prepOptions)
    B_est = SVAR_out['B_est']
    B_est_expected = np.array([[0.99661, 0.0, 0.0, 0.0, 0.0],
 [0.01134, 1.01706, 0.0, 0.0, 0.0],
 [0.00259, -0.03242, 0.98471, 0.00436, 0.01147],
 [-0.01471, -0.01624, 0.01191, 0.99008, 0.01359],
 [-0.01949, -0.00042, -0.00574, 0.00885, 0.98538]])

    B_est = np.round(B_est, 5)
    B_est_expected = np.round(B_est_expected, 5)

    assert B_est.tolist() == B_est_expected.tolist()

def test_GMM_BlockRecrusive(supply_eps):
    n = 5
    u = supply_eps[:, :n]

    block1 = np.array([1, 1])
    block2 = np.array([2, 2])
    block3 = np.array([3, 5])
    blocks = list()
    blocks.append(block1)
    blocks.append(block2)
    blocks.append(block3)

    prepOptions = dict()
    prepOptions['blocks'] = blocks
    prepOptions['printOutput'] = False
    SVAR_out = SVAR.SVARest(u, estimator='GMM_WF', prepOptions=prepOptions)
    B_est_PC = SVAR_out['B_est']

    prepOptions = dict()
    prepOptions['n_rec'] = 5
    prepOptions['Wstartopt'] = 'I'
    prepOptions['printOutput'] = False
    Rec_out = SVAR.SVARest(u, estimator='GMM', prepOptions=prepOptions)
    B_est_rec = Rec_out['B_est']

    B_est_rec = np.round(B_est_rec, 5)
    B_est_PC = np.round(B_est_PC, 5)

    assert  B_est_PC[:2,:2].tolist() == B_est_rec[:2,:2].tolist()

    expected = np.array([[0.003330608469677456, -0.033621407028708195],
 [-0.010995613002447795, -0.020220644924939964],
 [-0.0179660316466463, -0.0012376406712120119]] )

    B_est_PC = np.round(B_est_PC, 5)
    expected = np.round(expected, 5)

    assert B_est_PC[2:,:2].tolist() == expected.tolist()

    expected = np.array([[0.9845877900231426, 0.027117178672422547, 0.004666349227279161],
 [-0.003575857115028188, 1.0010856031042032, 0.0062304558880683265],
 [0.0020623518472946143, 0.010486623020412799, 0.988414549383155]])

    B_est_PC = np.round(B_est_PC, 5)
    expected = np.round(expected, 5)

    assert np.round(B_est_PC[2:, 2:],5).tolist() == np.round(expected,5).tolist()













import numpy as np
import statsmodels.api as sm
import matplotlib.pyplot as plt
import copy
from scipy.stats import norm
import SVAR
from tabulate import tabulate
import scipy.stats


# def var_statsmodel(y,lags):
#     model = VAR(y)
#     results = model.fit(lags)
#     print(results.summary())
#     return results
#
# def var_infocrit(y,maxlags,crit='aic'):
#     model = VAR(y)
#     model.select_order(maxlags)
#     results = model.fit(maxlags=maxlags, ic=crit)
#     print(results.summary())
#     return results



def get_ARMAtrices(coefmat, n, lags, add_const=True, add_trend=True, add_trend2=True):
    if add_const:
        const = coefmat[:, 0]
        coefmat = coefmat[:, 1:]
    else:
        const = np.zeros([n])
    if add_trend:
        trend = coefmat[:, 0]
        coefmat = coefmat[:, 1:]
    else:
        trend = np.zeros([n])
    if add_trend2:
        trend2 = coefmat[:, 0]
        coefmat = coefmat[:, 1:]
    else:
        trend2 = np.zeros([n])

    AR = np.zeros([n, n, lags])
    for i in range(lags):
        AR[:, :, i] = coefmat[:, :n]
        coefmat = coefmat[:, n:]

    return AR, const, trend, trend2





def OLS_ReducedForm(y, lags, add_const=True, add_trend=False, add_trend2=False, exog=[],showresults=False):
    T, n = y.shape

    shortY = y[lags:]

    X = np.zeros((T - lags, n * lags))
    for i in range(1, lags + 1):
        X[:, ( i-1) * n:( i ) * n] = y[lags - i:T - i]

    Xexo = np.ones((T - lags, 1)) if add_const else np.empty((T - lags, 0))

    if add_trend:
        Time = np.arange(T - lags).reshape(-1,1)
        Xexo = np.hstack((Xexo, Time))

    if add_trend2:
        Time = np.arange(T - lags).reshape(-1,1)
        Xexo = np.hstack((Xexo, np.power(Time,2)))

    X = np.hstack((Xexo, X))

    Atilde = np.linalg.lstsq(X, shortY, rcond=None)[0].T
    u = shortY - X @ Atilde.T

    AR, const, trend, trend2 = get_ARMAtrices(Atilde, n, lags, add_const=add_const, add_trend=add_trend,
                                              add_trend2=add_trend2)
    out = dict()
    out['u'] = u
    out['AR'] = AR
    out['const'] = const
    out['trend'] = trend
    out['trend2'] = trend2
    return out


def infocrit(y, maxLag, add_const=True, add_trend=False, add_trend2=False):
    T, n = np.shape(y)
    T = T - maxLag

    par_plus = int(add_const) + int(add_trend) + int(add_trend2)

    AIC_crit = np.full([maxLag + 1, 2], 0.)
    AIC_crit_row = np.full([n, maxLag + 1, 2], 0.)
    BIC_crit = np.full([maxLag + 1, 2], 0.)
    BIC_crit_row = np.full([n, maxLag + 1, 2], 0.)
    for this_lag in range(maxLag + 1):
        thisY = y[maxLag - this_lag:, :]

        out_redform = OLS_ReducedForm(thisY, this_lag, add_const=add_const, add_trend=add_trend,
                                      add_trend2=add_trend2)
        u = out_redform['u']
        this_aic = np.log(np.linalg.det(np.matmul(np.transpose(u), u) / T)) + 2 / T * (
                np.sum(this_lag) * np.power(n, 2) + par_plus * n)
        AIC_crit[this_lag, :] = np.array([this_lag, this_aic])

        this_BIC = np.log(np.linalg.det(np.matmul(np.transpose(u), u) / T)) + np.log(T) / T * (
                np.sum(this_lag) * np.power(n, 2) + par_plus * n)
        BIC_crit[this_lag, :] = np.array([this_lag, this_BIC])

        for i in range(n):
            this_aic_row = np.log(np.matmul(np.transpose(u[:, i]), u[:, i]) / T) + 2 / T * (
                    np.sum(this_lag) * n + par_plus)
            AIC_crit_row[i, this_lag, :] = np.array([this_lag, this_aic_row])

            this_bic_row = np.log(np.matmul(np.transpose(u[:, i]), u[:, i]) / T) + np.log(T) / T * (
                    np.sum(this_lag) * n + par_plus)
            BIC_crit_row[i, this_lag, :] = np.array([this_lag, this_bic_row])

    AIC_min_crit = np.argmin(AIC_crit[:, 1])
    BIC_min_crit = np.argmin(BIC_crit[:, 1])
    AIC_min_crit_row = np.zeros([n])
    BIC_min_crit_row = np.zeros([n])
    for i in range(n):
        AIC_min_crit_row[i] = np.argmin(AIC_crit_row[i, :, 1])
        BIC_min_crit_row[i] = np.argmin(BIC_crit_row[i, :, 1])

    out = dict()
    out['AIC_crit'] = AIC_crit
    out['AIC_crit_row'] = AIC_crit_row
    out['AIC_min_crit'] = AIC_min_crit
    out['AIC_min_crit_row'] = AIC_min_crit_row
    out['BIC_crit'] = BIC_crit
    out['BIC_crit_row'] = BIC_crit_row
    out['BIC_min_crit'] = BIC_min_crit
    out['BIC_min_crit_row'] = BIC_min_crit_row
    return out

def get_FEVD(u,B,AR,horizon):
    [T,n] = np.shape(u)

    e = SVAR.innovation(u, SVAR.get_BVector(B))
    scale = np.dot(np.transpose(e), e) / (T  - 1)
    B = np.dot(B, np.sqrt(np.diag(np.diag(scale))))

    [irf, phi] = SVAR.get_IRF(B, AR, irf_length=horizon, scale=False,outPhi=True)

    fevd = np.full([n,n],np.nan)
    for j in range(n):
        VarEj = np.sum(np.power(phi[j,:,:],2))
        for k in range(n):
            VarEjk = np.sum(np.power(phi[j,k,:],2))
            fevd[j,k] = VarEjk/VarEj

    return fevd

def get_IRF(B, AR, irf_length=12, scale=False, outPhi=False):
    n = np.shape(B)[0]

    if np.ndim(AR) == 2:
        lags = 1
        AR_new = np.full([n, n, 1], np.nan)
        AR_new[:, :, 0] = AR
        AR = AR_new
    else:
        lags = np.shape(AR)[2]
    phi = np.full([n, n, irf_length], np.nan)

    phi[:, :, 0] = np.eye(n, n)

    # normalize impact
    if scale:
        B = np.matmul(B, np.linalg.inv(np.diag(np.diag(B))))

    for i in range(1, irf_length):
        tmpsum = np.zeros([n, n])
        for j in range(i):
            if j < lags:
                tmpsum = tmpsum + np.matmul(phi[:, :, i - j - 1], AR[:, :, j])
        phi[:, :, i] = tmpsum

    for i in range(irf_length):
        phi[:, :, i] = np.matmul(phi[:, :, i], B)

    irf = np.full([irf_length, n, n], np.nan)
    for i in range(n):
        for j in range(irf_length):
            irf[j, :, i] = phi[:, i, j]
    if outPhi:
        out = irf, phi
    else:
        out = irf
    return out



def plot_IRF(irf, irf_bootstrap=[], alphas=np.array([0.2 / 2, 0.32 / 2]), shocks=[], responses=[], shocknames=[], responsnames=[], cummulative=[], sym=False):
    n = np.shape(irf)[1]
    if np.array(shocks).size == 0:
        shocks = np.arange(n)
    if np.array(responses).size == 0:
        responses = np.arange(n)
    if np.array(shocknames).size == 0:
        shocknames = np.arange(n)
    if np.array(responsnames).size == 0:
        responsnames = np.array([f"y_{{{i},t}}" for i in range(1, n + 1)])
    if np.array(cummulative).size == 0:
        cummulative = np.full(n, False)

    num_rows = np.size(shocks)
    num_cols = np.size(responses)
    fig_width = 6
    fig_height = fig_width * num_rows / num_cols


    fig, ax = plt.subplots(np.size(shocks), np.size(responses),   sharex=True,tight_layout=True, figsize=(fig_width, fig_height))


    plotcounter = 1

    for response_idx, response  in enumerate(responses):

        for shock_idx, shock in enumerate(shocks):
            y = irf[:, response, shock]
            try:
                y_bootstrap = np.asarray(irf_bootstrap)[:,:,response,shock]
                if cummulative[response]:
                    y_new = copy.deepcopy(y)
                    y_bootstrap_new = copy.deepcopy(y_bootstrap)
                    for t in range(np.size(y)):
                        y_new[t] = np.sum(y[0:t + 1])
                        y_bootstrap_new[:,t]= np.sum(y_bootstrap[:,0:t+1],axis=1)
                    y = copy.deepcopy(y_new)
                    y_bootstrap = y_bootstrap_new
            except:
                if cummulative[response]:
                    y_new = copy.deepcopy(y)
                    for t in range(np.size(y)):
                        y_new[t] = np.sum(y[0:t + 1])
                    y = copy.deepcopy(y_new)

            x = np.arange(np.shape(y)[0])


            ax[ shock_idx,response_idx].plot(x, y, '--',color="blue")
            ax[shock_idx,response_idx].axhline(0, color='black', linewidth=0.5, )
            # Plot Bootstrap irf quantiles
            for alpha in alphas:
                if np.array(irf_bootstrap).size != 0:
                    if sym:
                        alpha_n = 1 - alpha
                        s = np.sqrt(np.var(y_bootstrap, axis=0))
                        lower = y - norm.ppf(alpha_n) * s
                        upper = y + norm.ppf(alpha_n) * s

                    else:
                        lower = np.quantile(y_bootstrap, alpha, axis=0)
                        upper = np.quantile(y_bootstrap, 1 - alpha, axis=0)



                    ax[shock_idx,response_idx].fill_between(x, lower, upper, color='blue', alpha=0.25)

                ylab  = r'$\epsilon^{' + str(shocknames[ shock]) + '}$'
                xlab = r'$ ' + str(responsnames[response ]) + '$'

                if shock == shocks[-1]:
                    ax[shock_idx, response_idx  ].set(xlabel = xlab)
                if response == responses[0]:
                    ax[shock_idx, response_idx  ].set(ylabel= ylab)


            plotcounter += 1

    plt.show()

    return fig





def summarize_shocks(u,varnames):
    T,n = np.shape(u)
    TabData = list()
    TabData.append(np.append("Mean", np.round(np.mean(u, axis=0), 3)))
    TabData.append(np.append("Median", np.round(np.median(u, axis=0), 3)))
    TabData.append(np.append("Std. deviation", np.round(np.std(u, axis=0), 3)))
    TabData.append(np.append("Variance", np.round(np.var(u, axis=0), 3)))
    TabData.append(np.append("Skewness", np.round(scipy.stats.skew(u, axis=0), 3)))
    TabData.append(np.append("Kurtosis", np.round(scipy.stats.kurtosis(u, axis=0, fisher=False), 3)))
    JBTest = np.zeros(n)
    for i in range(n):
        JBTest[i] = np.round(scipy.stats.jarque_bera(u[:, i])[1], 3)
    TabData.append(np.append("Jarque-Bera Test (p-value)", JBTest))
    tableSummary = tabulate(TabData, headers=varnames, tablefmt="github")
    print(tableSummary)
    return TabData

def transform_SVARbootstrap_out(out_bootstrap):
    out_boootstrap_irf = np.full([len(out_bootstrap),
                                  np.shape(out_bootstrap[0][2])[0],
                                  np.shape(out_bootstrap[0][2])[1],
                                  np.shape(out_bootstrap[0][2])[2]], 0.0)
    out_bootstrap_B = np.full([len(out_bootstrap),
                               np.shape(out_bootstrap[0][1])[0],
                               np.shape(out_bootstrap[0][1])[1]], 0.0)
    out_bootstrap_AR = np.full([len(out_bootstrap),
                                np.shape(out_bootstrap[0][0])[0],
                                np.shape(out_bootstrap[0][0])[1],
                                np.shape(out_bootstrap[0][0])[2]], 0.0)
    for i in range(len(out_bootstrap)):
        out_boootstrap_irf[i, :, :, :] = out_bootstrap[i][2]
        out_bootstrap_B[i, :, :] = out_bootstrap[i][1]
        out_bootstrap_AR[i, :, :, :] = out_bootstrap[i][0]

    return out_boootstrap_irf, out_bootstrap_B, out_bootstrap_AR


import numpy as np


def MoG_rnd(Omega, lamb, T):
    # MoG random variable

    select = np.random.uniform(0, 1, T)
    mu1 = Omega[0, 0]
    sigma1 = np.sqrt(Omega[0, 1])
    x1 = np.random.normal(loc=mu1, scale=sigma1, size=T)
    mu2 = Omega[1, 0]
    sigma2 = np.sqrt(Omega[1, 1])
    x2 = np.random.normal(loc=mu2, scale=sigma2, size=T)

    x = x1
    x[select >= lamb] = x2[select >= lamb]
    return x


def MoG_Moments(Omega, lamb):
    # Calculates first six moments of MoG

    mu1, w11 = Omega[0, :]
    mu2, w22 = Omega[1, :]

    sigma1 = np.sqrt(w11)
    sigma2 = np.sqrt(w22)

    m0 = 1

    # m1_o = lamb * mu1 + (1 - lamb) * mu2
    # m2_o = (lamb) * (np.power(mu1, 2) + w11) \
    #      + (1 - lamb) * (np.power(mu2, 2) + w22)
    # m3_o = (lamb) * (np.power(mu1, 3) + 3 * mu1 * w11) \
    #      + (1 - lamb) * (np.power(mu2, 3) + 3 * mu2 * w22)
    # m4_o = (lamb) * (np.power(mu1, 4) + 6 * np.power(mu1, 2) * w11 + 3 * np.power(w11, 2)) \
    #      + (1 - lamb) * (np.power(mu2, 4) + 6 * np.power(mu2, 2) * w22 + 3 * np.power(w22, 2))
    # m5_o = (lamb) * (np.power(mu1, 5) + 10 * np.power(mu1, 3) * w11 + 15 * mu1 * np.power(w11, 2)) \
    #      + (1 - lamb) * (np.power(mu2, 5) + 10 * np.power(mu2, 3) * w22 + 15 * mu2 * np.power(w22, 2))
    # m6_o = (lamb) * (np.power(mu1, 6) + 15 * np.power(mu1, 4) * w11 + 45 * np.power(mu1, 2) * np.power(w11,
    #                                                                                                  2) + 15 * np.power(
    #     w11, 3)) \
    #      + (1 - lamb) * (np.power(mu2, 6) + 15 * np.power(mu2, 4) * w22 + 45 * np.power(mu2, 2) * np.power(w22,
    #                                                                                                        2) + 15 * np.power(
    #     w22, 3))

    m1 = lamb * mu1 + (1 - lamb) * mu2
    m2 = (lamb) * (np.power(mu1, 2) + np.power(sigma1, 2)) \
         + (1 - lamb) * (np.power(mu2, 2) + np.power(sigma2, 2))
    m3 = (lamb) * (np.power(mu1, 3) + 3 * mu1 * np.power(sigma1, 2)) \
         + (1 - lamb) * (np.power(mu2, 3) + 3 * mu2 * np.power(sigma2, 2))
    m4 = (lamb) *           (np.power(mu1, 4) +
                           6 * np.power(mu1, 2) * np.power(sigma1, 2) +
                           3 * np.power(sigma1, 4)) \
         + (1 - lamb) *     (np.power(mu2, 4) +
                             6 * np.power(mu2, 2) * np.power(sigma2, 2) +
                             3 * np.power(sigma2, 4))

    m5 = (lamb) * (np.power(mu1, 5) +
                   10 * np.power(mu1, 3) * np.power(sigma1, 2) +
                   15 * mu1 * np.power(sigma1, 4)) \
         + (1 - lamb) * (np.power(mu2, 5) +
                         10 * np.power(mu2, 3) * np.power(sigma2, 2) +
                         15 * mu2 * np.power(sigma2, 4))
    m6 = (lamb) * (np.power(mu1, 6) +
                   15 * np.power(mu1, 4) * np.power(sigma1, 2) +
                   45 * np.power(mu1, 2) * np.power(  sigma1, 4) +
                   15 * np.power(sigma1, 6)) \
         + (1 - lamb) * (np.power(mu2, 6) +
                         15 * np.power(mu2, 4) * np.power(sigma2, 2) +
                         45 * np.power(mu2, 2) * np.power( sigma2, 4) +
                         15 * np.power(sigma2, 6))
    moments = np.array([m0, m1, m2, m3, m4, m5, m6])
    return moments


def MoG_Cumulants(moments):
    m0, m1, m2, m3, m4, m5, m6 = moments
    k1 = m1
    k2 = m2 - np.power(m1, 2)
    k3 = m3 - 3 * m1 * m2 + 2 * np.power(m1, 3)
    k4 = m4 - 4 * m1 * m3 - 3 * np.power(m2, 2) + 12 * np.power(m1, 2) * m2 - 6 * np.power(m1, 4)
    k5 = m5 - 5 * m1 * m4 - 10 * m2 * m3 + 20 * np.power(m1, 2) * m3 + 30 * m1 * np.power(m2, 2) - 60 * np.power(m1,
                                                                                                                 3) * m2 + 24 * np.power(
        m1, 5)
    k6 = m6 - 6 * m1 * m5 - 15 * m2 * m4 + 30 * np.power(m1, 2) * m4 - 10 * np.power(m3,
                                                                                     2) + 120 * m1 * m2 * m3 - 120 * np.power(
        m1, 3) * m3 + 30 * np.power(m2, 3) - 270 * np.power(m1, 2) * np.power(m2, 2) + 360 * np.power(m1,
                                                                                                      4) * m2 - 120 * np.power(
        m1, 6)

    return k1, k2, k3, k4, k5, k6


def MoG_Get_MoG(moments):
    k1, k2, k3, k4, k5, k6 = MoG_Cumulants(moments)

    poly = lambda p: (8) * np.power(p, 9) + \
                     (28 * k4) * np.power(p, 7) + \
                     (12 * np.power(k3, 2)) * np.power(p, 6) + \
                     (24 * k3 * k5 + 30 * np.power(k4, 2)) * np.power(p, 5) + \
                     (148 * np.power(k3, 2) * k4 - 6 * np.power(k5, 2)) * np.power(p, 4) + \
                     (96 * np.power(k3, 4) + 9 * np.power(k4, 3) - 36 * k3 * k4 * k5) * np.power(p, 3) + \
                     (-21 * np.power(k3, 2) * np.power(k4, 2) - 24 * np.power(k3, 3) * k5) * np.power(p, 2) + \
                     (-32 * np.power(k3, 4) * k4) * np.power(p, 1) + \
                     (-8 * np.power(k3, 6))

    pol = np.array([(8), 0, (28 * k4), (12 * np.power(k3, 2)), (24 * k3 * k5 + 30 * np.power(k4, 2)),
                    (148 * np.power(k3, 2) * k4 - 6 * np.power(k5, 2)),
                    (96 * np.power(k3, 4) + 9 * np.power(k4, 3) - 36 * k3 * k4 * k5),
                    (-21 * np.power(k3, 2) * np.power(k4, 2) - 24 * np.power(k3, 3) * k5), (-32 * np.power(k3, 4) * k4),
                    (-8 * np.power(k3, 6))])
    p = np.roots(pol)  # find roots of polynomial
    p = p[~np.iscomplex(p)]  # only real roots
    p = p.real
    p = p[p <= 0]

    for this_p in p:
        print(this_p / mu1)
        if (4 * np.power(this_p, 3) * k3 - 4 * np.power(k3, 3) - 6 * this_p * k3 * k4 - 2 * np.power(this_p,
                                                                                                     2) * k5) != 0:
            this_s = - (4 * np.power(this_p, 5) + 14 * np.power(this_p, 2) * np.power(k3, 2) + 8 * np.power(this_p,
                                                                                                            3) * k4 + np.power(
                k3, 2) * k4 + 3 * this_p * np.power(k4, 2) - 2 * this_p * k3 * k5) / (
                                 4 * np.power(this_p, 3) * k3 - 4 * np.power(k3,
                                                                             3) - 6 * this_p * k3 * k4 - 2 * np.power(
                             this_p, 2) * k5)
            poly = np.array([1, -this_s, this_p])
            mu1, mu2 = np.roots(poly)
            print("mu1: ", mu1, " mu2: ", mu2)
            lamb = -mu2 / (mu1 - mu2)
            print("Lambda: ", lamb)

            R1 = -k2 - this_p
            R1_tilde = R1 * (mu1 - mu2)
            R2 = (this_p * this_p - k3) / (3 * this_p)
            R2_tilde = R2 * (mu1 - mu2)

            sigma1 = np.sqrt((R1_tilde - mu1 * R2_tilde) / (mu1 - mu2))
            sigma2 = np.sqrt(R2_tilde + np.power(sigma1, 2))
            print("sigma1: ", sigma1, " sigma2: ", sigma2)



from SVAR.estSVAR import SVARest
from SVAR.estimatorCholesky import get_B_Cholesky
from SVAR.SVARbasics import get_IRF, plot_IRF, OLS_ReducedForm
from SVAR.estSVARbootstrap import bootstrap_SVAR
from SVAR.SVARutil import innovation, get_BMatrix, get_BVector, do_whitening



def test():
    print("Package: SVAR-GMM is working")
